# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2020 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
import pytest
import numpy
from copy import deepcopy
from datacube.testutils import mk_sample_dataset, mk_sample_product
from datacube.model import (DatasetType, GridSpec, Measurement,
                            MetadataType, Range, ranges_overlap)
from datacube.utils import geometry
from datacube.utils.documents import InvalidDocException
from datacube.storage import measurement_paths
from datacube.testutils.geom import AlbersGS
from datacube.api.core import output_geobox


def test_gridspec():
    gs = GridSpec(crs=geometry.CRS('EPSG:4326'), tile_size=(1, 1), resolution=(-0.1, 0.1), origin=(10, 10))
    poly = geometry.polygon([(10, 12.2), (10.8, 13), (13, 10.8), (12.2, 10), (10, 12.2)], crs=geometry.CRS('EPSG:4326'))
    cells = {index: geobox for index, geobox in list(gs.tiles_from_geopolygon(poly))}
    assert set(cells.keys()) == {(0, 1), (0, 2), (1, 0), (1, 1), (1, 2), (2, 0), (2, 1)}
    assert numpy.isclose(cells[(2, 0)].coordinates['longitude'].values, numpy.linspace(12.05, 12.95, num=10)).all()
    assert numpy.isclose(cells[(2, 0)].coordinates['latitude'].values, numpy.linspace(10.95, 10.05, num=10)).all()

    # check geobox_cache
    cache = {}
    poly = gs.tile_geobox((3, 4)).extent
    (c1, gbox1), = list(gs.tiles_from_geopolygon(poly, geobox_cache=cache))
    (c2, gbox2), = list(gs.tiles_from_geopolygon(poly, geobox_cache=cache))

    assert c1 == (3, 4) and c2 == c1
    assert gbox1 is gbox2

    assert '4326' in str(gs)
    assert '4326' in repr(gs)
    assert (gs == gs)
    assert (gs == {}) is False


def test_gridspec_upperleft():
    """ Test to ensure grid indexes can be counted correctly from bottom left or top left
    """
    tile_bbox = geometry.BoundingBox(left=1934400.0, top=2414800.0, right=2084400.000, bottom=2264800.000)
    bbox = geometry.BoundingBox(left=1934615, top=2379460, right=1937615, bottom=2376460)
    # Upper left - validated against WELD product tile calculator
    # http://globalmonitoring.sdstate.edu/projects/weld/tilecalc.php
    gs = GridSpec(crs=geometry.CRS('EPSG:5070'), tile_size=(-150000, 150000), resolution=(-30, 30),
                  origin=(3314800.0, -2565600.0))
    cells = {index: geobox for index, geobox in list(gs.tiles(bbox))}
    assert set(cells.keys()) == {(30, 6)}
    assert cells[(30, 6)].extent.boundingbox == tile_bbox

    gs = GridSpec(crs=geometry.CRS('EPSG:5070'), tile_size=(150000, 150000), resolution=(-30, 30),
                  origin=(14800.0, -2565600.0))
    cells = {index: geobox for index, geobox in list(gs.tiles(bbox))}
    assert set(cells.keys()) == {(30, 15)}  # WELD grid spec has 21 vertical cells -- 21 - 6 = 15
    assert cells[(30, 15)].extent.boundingbox == tile_bbox


def test_dataset_basics():
    ds = mk_sample_dataset([dict(name='a')])
    assert ds == ds
    assert ds != "33"
    assert (ds == "33") is False
    assert str(ds) == repr(ds)

    ds = mk_sample_dataset([dict(name='a')], uri=None, geobox=None)
    assert ds.uris == []
    assert ds.uri_scheme == ''
    assert ds.crs is None
    assert ds.bounds is None
    assert ds.extent is None
    assert ds.transform is None


def test_dataset_measurement_paths():
    format = 'GeoTiff'

    ds = mk_sample_dataset([dict(name=n,
                                 path=n+'.tiff')
                            for n in 'a b c'.split(' ')],
                           uri='file:///tmp/datataset.yml',
                           format=format)

    assert ds.local_uri == ds.uris[0]
    assert ds.uri_scheme == 'file'
    assert ds.format == format
    paths = measurement_paths(ds)

    for k, v in paths.items():
        assert v == 'file:///tmp/' + k + '.tiff'

    ds.uris = None
    assert ds.local_uri is None
    with pytest.raises(ValueError):
        measurement_paths(ds)


def test_product_basics():
    product = mk_sample_product('test_product')
    assert product.name == 'test_product'
    assert 'test_product' in str(product)
    assert 'test_product' in repr(product)
    assert product == product
    assert product == mk_sample_product('test_product')
    assert not (product == mk_sample_product('other'))
    assert not (product == [()])
    assert hash(product) == hash(mk_sample_product('test_product'))
    assert 'time' in dir(product.metadata)

    assert product.measurements == product.lookup_measurements()
    assert product.lookup_measurements(['red']) == product.lookup_measurements('red')


def test_product_dimensions():
    product = mk_sample_product('test_product')
    assert product.grid_spec is None
    assert product.dimensions == ('time', 'y', 'x')

    product = mk_sample_product('test_product', with_grid_spec=True)
    assert product.grid_spec is not None
    assert product.dimensions == ('time', 'y', 'x')

    partial_storage = product.definition['storage']
    partial_storage.pop('tile_size')
    product = mk_sample_product('tt', storage=partial_storage)
    assert product.grid_spec is None


def test_product_nodata_nan():
    # When storing .nan to JSON in DB it becomes a string with value "NaN"
    # Make sure it is converted back to real NaN
    product = mk_sample_product('test', measurements=[dict(name='_nan',
                                                           dtype='float32',
                                                           nodata='NaN'),
                                                      dict(name='_inf',
                                                           dtype='float32',
                                                           nodata='Infinity'),
                                                      dict(name='_neg_inf',
                                                           dtype='float32',
                                                           nodata='-Infinity'),
                                                      ])
    for m in product.measurements.values():
        assert isinstance(m.nodata, float)

    assert numpy.isnan(product.measurements['_nan'].nodata)
    assert product.measurements['_inf'].nodata == numpy.inf
    assert product.measurements['_neg_inf'].nodata == -numpy.inf


def test_product_scale_factor():
    product = mk_sample_product('test', measurements=[dict(name='red',
                                                           scale_factor=33,
                                                           add_offset=-5)])
    assert product.validate(product.definition) is None
    assert product.measurements['red'].scale_factor == 33
    assert product.measurements['red'].add_offset == -5
    attrs = product.measurements['red'].dataarray_attrs()
    assert attrs['scale_factor'] == 33
    assert attrs['add_offset'] == -5


def test_product_load_hints():
    product = mk_sample_product('test_product',
                                load=dict(crs='epsg:3857',
                                          resolution={'x': 10, 'y': -10}))

    assert 'load' in product.definition
    assert DatasetType.validate(product.definition) is None

    hints = product._extract_load_hints()
    assert hints['crs'] == geometry.CRS('epsg:3857')
    assert hints['resolution'] == (-10, 10)
    assert 'align' not in hints

    product = mk_sample_product('test_product',
                                load=dict(crs='epsg:3857',
                                          align={'x': 5, 'y': 6},
                                          resolution={'x': 10, 'y': -10}))

    hints = product.load_hints()
    assert hints['output_crs'] == geometry.CRS('epsg:3857')
    assert hints['resolution'] == (-10, 10)
    assert hints['align'] == (6, 5)
    assert product.default_crs == geometry.CRS('epsg:3857')
    assert product.default_resolution == (-10, 10)
    assert product.default_align == (6, 5)

    product = mk_sample_product('test_product',
                                load=dict(crs='epsg:4326',
                                          align={'longitude': 0.5, 'latitude': 0.6},
                                          resolution={'longitude': 1.2, 'latitude': -1.1}))

    hints = product.load_hints()
    assert hints['output_crs'] == geometry.CRS('epsg:4326')
    assert hints['resolution'] == (-1.1, 1.2)
    assert hints['align'] == (0.6, 0.5)

    # check it's cached
    assert product.load_hints() is product.load_hints()

    # check schema: crs and resolution are compulsory
    for k in ('resolution', 'crs'):
        doc = deepcopy(product.definition)
        assert DatasetType.validate(doc) is None

        doc['load'].pop(k)
        assert k not in doc['load']

        with pytest.raises(InvalidDocException):
            DatasetType.validate(doc)

    # check GridSpec leakage doesn't happen for fully defined gridspec
    product = mk_sample_product('test', with_grid_spec=True)
    assert product.grid_spec is not None
    assert product.load_hints() == {}

    # check for fallback into partially defined `storage:`
    product = mk_sample_product('test', storage=dict(
        crs='EPSG:3857',
        resolution={'x': 10, 'y': -10}))
    assert product.grid_spec is None
    assert product.default_resolution == (-10, 10)
    assert product.default_crs == geometry.CRS('EPSG:3857')

    # check for fallback into partially defined `storage:`
    # no resolution -- no hints
    product = mk_sample_product('test', storage=dict(
        crs='EPSG:3857'))
    assert product.grid_spec is None
    assert product.load_hints() == {}

    # check misspelled load hints
    product = mk_sample_product('test_product',
                                load=dict(crs='epsg:4326',
                                          resolution={'longtude': 1.2, 'latitude': -1.1}))
    assert product.load_hints() == {}


def test_measurement():
    # Can create a measurement
    m = Measurement(name='t', dtype='uint8', nodata=255, units='1')

    # retrieve it's vital stats
    assert m.name == 't'
    assert m.dtype == 'uint8'
    assert m.nodata == 255
    assert m.units == '1'

    # retrieve the information required for filling a DataArray
    assert m.dataarray_attrs() == {'nodata': 255, 'units': '1'}

    # Can add a new attribute by name and ensure it updates the DataArray attrs too
    m['bob'] = 10
    assert m.bob == 10
    assert m.dataarray_attrs() == {'nodata': 255, 'units': '1', 'bob': 10}

    m['none'] = None
    assert m.none is None

    # Resampling method is special and *not* needed for DataArray attrs
    m['resampling_method'] = 'cubic'
    assert 'resampling_method' not in m.dataarray_attrs()

    # It's possible to copy and update a Measurement instance
    m2 = m.copy()
    assert m2.bob == 10
    assert m2.dataarray_attrs() == m.dataarray_attrs()

    assert repr(m2) == repr(m)

    # Must specify *all* required keys. name, dtype, nodata and units
    with pytest.raises(ValueError) as e:
        Measurement(name='x', units='1', nodata=0)

    assert 'required keys missing:' in str(e.value)
    assert 'dtype' in str(e.value)


def test_output_geobox_load_hints():
    geobox0 = AlbersGS.tile_geobox((15, -40))

    geobox = output_geobox(load_hints={'output_crs': geobox0.crs,
                                       'resolution': geobox0.resolution},
                           geopolygon=geobox0.extent)
    assert geobox == geobox0


def test_like_geobox():
    geobox = AlbersGS.tile_geobox((15, -40))
    assert output_geobox(like=geobox) is geobox


def test_output_geobox_fail_paths():

    with pytest.raises(ValueError):
        output_geobox()

    with pytest.raises(ValueError):
        output_geobox(output_crs='EPSG:4326')  # need resolution as well

    # need bounds
    with pytest.raises(ValueError):
        output_geobox(output_crs='EPSG:4326', resolution=(1, 1))


def test_metadata_type():
    m = MetadataType({'name': 'eo',
                      'dataset': dict(
                          id=['id'],
                          label=['ga_label'],
                          creation_time=['creation_dt'],
                          measurements=['image', 'bands'],
                          sources=['lineage', 'source_datasets'],
                          format=['format', 'name'])},
                     dataset_search_fields={})

    assert 'eo' in str(m)
    assert 'eo' in repr(m)
    assert m.name == 'eo'
    assert m.description is None
    assert m.dataset_reader({}) is not None

    # again but without dataset_search_fields
    m = MetadataType(m.definition)
    assert 'eo' in str(m)
    assert 'eo' in repr(m)
    assert m.name == 'eo'
    assert m.description is None
    assert m.dataset_reader({}) is not None


def test_ranges_overlap():
    assert not ranges_overlap(
        Range(begin=1, end=5),
        Range(begin=11, end=15)
    )
    assert not ranges_overlap(
        Range(begin=1, end=5),
        Range(begin=5, end=11)
    )
    assert not ranges_overlap(
        Range(begin=5, end=11),
        Range(begin=1, end=5)
    )
    assert ranges_overlap(
        Range(begin=1, end=5),
        Range(begin=1, end=5)
    )
    assert ranges_overlap(
        Range(begin=1, end=15),
        Range(begin=10, end=25)
    )
    assert ranges_overlap(
        Range(begin=10, end=25),
        Range(begin=1, end=15)
    )
    assert ranges_overlap(
        Range(begin=1, end=25),
        Range(begin=10, end=15)
    )
    assert ranges_overlap(
        Range(begin=10, end=15),
        Range(begin=1, end=25)
    )



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2020 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
"""
Common datatypes for DB drivers.
"""

from datetime import date, datetime, time
from dateutil.tz import tz
from typing import List

from datacube.model import Range
from datacube.model.fields import Expression, Field

__all__ = ['Field',
           'Expression',
           'OrExpression',
           'UnknownFieldError',
           'to_expressions',
           'as_expression']


class UnknownFieldError(Exception):
    pass


class OrExpression(Expression):
    def __init__(self, *exprs):
        super(OrExpression, self).__init__()
        self.exprs = exprs
        # Or expressions built by dc.load are always made up of simple expressions that share the same field.
        self.field = exprs[0].field

    def evaluate(self, ctx):
        return any(expr.evaluate(ctx) for expr in self.exprs)


def as_expression(field: Field, value) -> Expression:
    """
    Convert a single field/value to expression, following the "simple" convensions.
    """
    if isinstance(value, Range):
        return field.between(value.begin, value.end)
    elif isinstance(value, list):
        return OrExpression(*(as_expression(field, val) for val in value))
    # Treat a date (day) as a time range.
    elif isinstance(value, date) and not isinstance(value, datetime):
        return as_expression(
            field,
            Range(
                datetime.combine(value, time.min.replace(tzinfo=tz.tzutc())),
                datetime.combine(value, time.max.replace(tzinfo=tz.tzutc()))
            )
        )
    return field == value


def _to_expression(get_field, name: str, value) -> Expression:
    field = get_field(name)
    if field is None:
        raise UnknownFieldError('Unknown field %r' % name)

    return as_expression(field, value)


def to_expressions(get_field, **query) -> List[Expression]:
    """
    Convert a simple query (dict of param names and values) to expression objects.
    :type get_field: (str) -> Field
    :type query: dict[str,str|float|datacube.model.Range]
    """
    return [_to_expression(get_field, name, value) for name, value in query.items()]



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2020 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
from typing import List, Optional

from ._tools import singleton_setup
from .driver_cache import load_drivers
from ..index.abstract import AbstractIndexDriver


class IndexDriverCache(object):
    def __init__(self, group: str) -> None:
        self._drivers = load_drivers(group)

        if len(self._drivers) == 0:
            from datacube.index.postgres.index import index_driver_init
            self._drivers = dict(default=index_driver_init())

        for driver in list(self._drivers.values()):
            if hasattr(driver, 'aliases'):
                for alias in driver.aliases:
                    self._drivers[alias] = driver

    def __call__(self, name: str) -> AbstractIndexDriver:
        """
        :returns: None if driver with a given name is not found

        :param name: Driver name
        :return: Returns IndexDriver
        """
        return self._drivers.get(name, None)

    def drivers(self) -> List[str]:
        """ Returns list of driver names
        """
        return list(self._drivers.keys())


def index_cache() -> IndexDriverCache:
    """ Singleton for IndexDriverCache
    """
    return singleton_setup(index_cache, '_instance',
                           IndexDriverCache,
                           'datacube.plugins.index')


def index_drivers() -> List[str]:
    """ Returns list driver names
    """
    return index_cache().drivers()


def index_driver_by_name(name: str) -> Optional[AbstractIndexDriver]:
    """ Lookup writer driver by name

    :returns: Initialised writer driver instance
    :returns: None if driver with this name doesn't exist
    """
    return index_cache()(name)



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2020 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
"""
Tools for masking data based on a bit-mask variable with attached definition.

The main functions are `make_mask(variable)` `describe_flags(variable)`
"""

import collections

import pandas
import xarray
from xarray import DataArray, Dataset

from datacube.utils.math import valid_mask


FLAGS_ATTR_NAME = 'flags_definition'


def list_flag_names(variable):
    """
    Returns the available masking flags for the variable

    :param variable: Masking xarray.Dataset or xarray.DataArray
    :return: list
    """
    flags_def = get_flags_def(variable)
    return sorted(list(flags_def.keys()))


def describe_variable_flags(variable, with_pandas=True):
    """
    Returns either a Pandas Dataframe (with_pandas=True - default) or a string
    (with_pandas=False) describing the available flags for a masking variable

    Interprets the `flags_definition` attribute on the provided variable and
    returns a Pandas Dataframe or string like::

        Bits are listed from the MSB (bit 13) to the LSB (bit 0)
        Bit     Value   Flag Name            Description
        13      0       cloud_shadow_fmask   Cloud Shadow (Fmask)
        12      0       cloud_shadow_acca    Cloud Shadow (ACCA)
        11      0       cloud_fmask          Cloud (Fmask)
        10      0       cloud_acca           Cloud (ACCA)

    :param variable: Masking xarray.Dataset or xarray.DataArray
    :return: Pandas Dataframe or str
    """
    flags_def = get_flags_def(variable)

    if not with_pandas:
        return describe_flags_def(flags_def)

    return pandas.DataFrame.from_dict(flags_def, orient='index')


def describe_flags_def(flags_def):
    return '\n'.join(generate_table(list(_table_contents(flags_def))))


def _table_contents(flags_def):
    yield 'Flag name', 'Description', 'Bit. No', 'Value', 'Meaning'
    for name, defn in sorted(flags_def.items(), key=_order_bitdefs_by_bits):
        name, desc = name, defn['description']
        for value, meaning in defn['values'].items():
            yield name, desc, str(defn['bits']), str(value), str(meaning)
            name, desc = '', ''


def _order_bitdefs_by_bits(bitdef):
    name, defn = bitdef
    try:
        return min(defn['bits'])
    except TypeError:
        return defn['bits']


def make_mask(variable, **flags):
    """
    Returns a mask array, based on provided flags

    When multiple flags are provided, they will be combined in a logical AND fashion.

    For example:

    >>> make_mask(pqa, cloud_acca=False, cloud_fmask=False, land_obs=True) # doctest: +SKIP

    OR

    >>> make_mask(pqa, **GOOD_PIXEL_FLAGS) # doctest: +SKIP

    where `GOOD_PIXEL_FLAGS` is a dict of flag_name to True/False

    :param variable:
    :type variable: xarray.Dataset or xarray.DataArray
    :param flags: list of boolean flags
    :return: boolean xarray.DataArray or xarray.Dataset
    """
    flags_def = get_flags_def(variable)

    mask, mask_value = create_mask_value(flags_def, **flags)

    return variable & mask == mask_value


def valid_data_mask(data):
    """
    Returns bool arrays where the data is not `nodata`

    :param Dataset or DataArray data:
    :return: Dataset or DataArray
    """
    if isinstance(data, Dataset):
        return data.map(valid_data_mask)

    if not isinstance(data, DataArray):
        raise TypeError('valid_data_mask not supported for type {}'.format(type(data)))

    nodata = data.attrs.get('nodata', None)

    return xarray.apply_ufunc(valid_mask, data, nodata,
                              dask='parallelized',
                              output_dtypes=[bool])


def mask_invalid_data(data, keep_attrs=True):
    """
    Sets all `nodata` values to ``nan``.

    This will convert numeric data to type `float`.

    :param Dataset or DataArray data:
    :param bool keep_attrs: If the attributes of the data should be included in the returned .
    :return: Dataset or DataArray
    """
    if isinstance(data, Dataset):
        # Pass keep_attrs as a positional arg to the DataArray func
        return data.map(mask_invalid_data, keep_attrs=keep_attrs, args=(keep_attrs,))

    if isinstance(data, DataArray):
        if 'nodata' not in data.attrs:
            return data
        out_data_array = data.where(data != data.nodata)
        if keep_attrs:
            out_data_array.attrs = {key: value
                                    for key, value in data.attrs.items()
                                    if key != 'nodata'}
        return out_data_array

    raise TypeError('mask_invalid_data not supported for type {}'.format(type(data)))


def create_mask_value(bits_def, **flags):
    mask = 0
    value = 0

    for flag_name, flag_ref in flags.items():
        defn = bits_def.get(flag_name, None)
        if defn is None:
            raise ValueError(f'Unknown flag: "{flag_name}"')

        try:
            [flag_value] = (bit_val
                            for bit_val, val_ref in defn['values'].items()
                            if val_ref == flag_ref)
            flag_value = int(flag_value)  # Might be string if coming from DB
        except ValueError:
            raise ValueError('Unknown value %s specified for flag %s' %
                             (flag_ref, flag_name))

        if isinstance(defn['bits'], collections.abc.Iterable):  # Multi-bit flag
            # Set mask
            for bit in defn['bits']:
                mask = set_value_at_index(mask, bit, True)

            shift = min(defn['bits'])
            real_val = flag_value << shift

            value |= real_val

        else:
            bit = defn['bits']
            mask = set_value_at_index(mask, bit, True)
            value = set_value_at_index(value, bit, flag_value)

    return mask, value


def mask_to_dict(bits_def, mask_value):
    """
    Describes which flags are set for a mask value

    :param bits_def:
    :param mask_value:
    :return: Mapping of flag_name -> set_value
    :rtype: dict
    """
    return_dict = {}
    for flag_name, flag_defn in bits_def.items():

        # Make bits a list, even if there is only one
        flag_bits = flag_defn['bits']
        if not isinstance(flag_defn['bits'], list):
            flag_bits = [flag_bits]

        # The amount to shift flag_value to line up with mask_value
        flag_shift = min(flag_bits)

        # Mask our mask_value, we are only interested in the bits for this flag
        flag_mask = 0
        for i in flag_bits:
            flag_mask |= (1 << i)
        masked_mask_value = mask_value & flag_mask

        for flag_value, value in flag_defn['values'].items():
            shifted_value = int(flag_value) << flag_shift
            if shifted_value == masked_mask_value:
                assert flag_name not in return_dict
                return_dict[flag_name] = value
    return return_dict


def get_flags_def(variable):
    flags = getattr(variable, FLAGS_ATTR_NAME, None)
    if flags is not None:
        return flags

    data_vars = getattr(variable, 'data_vars', None)

    if data_vars is not None:
        # Maybe we have a DataSet, not a DataArray
        for var in data_vars.values():
            flags = getattr(var, FLAGS_ATTR_NAME, None)
            if flags is not None:
                return flags

    raise ValueError('No masking variable found')


def set_value_at_index(bitmask, index, value):
    """
    Set a bit value onto an integer bitmask

    eg. set bits 2 and 4 to True
    >>> mask = 0
    >>> mask = set_value_at_index(mask, 2, True)
    >>> mask = set_value_at_index(mask, 4, True)
    >>> print(bin(mask))
    0b10100
    >>> mask = set_value_at_index(mask, 2, False)
    >>> print(bin(mask))
    0b10000

    :param bitmask: existing int bitmask to alter
    :type bitmask: int
    :type index: int
    :type value: bool
    """
    bit_val = 2 ** index
    if value:
        bitmask |= bit_val
    else:
        bitmask &= (~bit_val)
    return bitmask


def generate_table(rows):
    """
    Yield strings to print a table using the data in `rows`.

    TODO: Maybe replace with Pandas

    :param rows: A sequence of sequences with the 0th element being the table
                 header
    """

    # - figure out column widths
    widths = [len(max(columns, key=len)) for columns in zip(*rows)]

    # - print the header
    header, data = rows[0], rows[1:]
    yield (
        ' | '.join(format(title, "%ds" % width) for width, title in zip(widths, header))
    )

    # Print the separator
    first_col = ''
    # - print the data
    for row in data:
        if first_col == '' and row[0] != '':
            # - print the separator
            yield '-+-'.join('-' * width for width in widths)
        first_col = row[0]

        yield (
            " | ".join(format(cdata, "%ds" % width) for width, cdata in zip(widths, row))
        )



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2020 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
import logging

from cachetools.func import lru_cache

from datacube.index.abstract import AbstractMetadataTypeResource
from datacube.index.postgis._transaction import IndexResourceAddIn
from datacube.model import MetadataType
from datacube.utils import jsonify_document, changes, _readable_offset
from datacube.utils.changes import check_doc_unchanged, get_doc_changes

_LOG = logging.getLogger(__name__)


class MetadataTypeResource(AbstractMetadataTypeResource, IndexResourceAddIn):
    def __init__(self, db, index):
        """
        :type db: datacube.drivers.postgis._connections.PostgresDb
        """
        self._db = db
        self._index = index

        self.get_unsafe = lru_cache()(self.get_unsafe)
        self.get_by_name_unsafe = lru_cache()(self.get_by_name_unsafe)

    def __getstate__(self):
        """
        We define getstate/setstate to avoid pickling the caches
        """
        return (self._db,)

    def __setstate__(self, state):
        """
        We define getstate/setstate to avoid pickling the caches
        """
        self.__init__(*state)

    def from_doc(self, definition):
        """
        :param dict definition:
        :rtype: datacube.model.MetadataType
        """
        MetadataType.validate(definition)
        return self._make(definition)

    def add(self, metadata_type, allow_table_lock=False):
        """
        :param datacube.model.MetadataType metadata_type:
        :param allow_table_lock:
            Allow an exclusive lock to be taken on the table while creating the indexes.
            This will halt other user's requests until completed.

            If false, creation will be slightly slower and cannot be done in a transaction.
        :rtype: datacube.model.MetadataType
        """
        # This column duplication is getting out of hand:
        MetadataType.validate(metadata_type.definition)

        existing = self.get_by_name(metadata_type.name)
        if existing:
            # They've passed us the same one again. Make sure it matches what is stored.
            _LOG.warning(f"Metadata Type {metadata_type.name} is already in the database, checking for differences")
            check_doc_unchanged(
                existing.definition,
                jsonify_document(metadata_type.definition),
                'Metadata Type {}'.format(metadata_type.name)
            )
        else:
            with self._db_connection() as connection:
                connection.insert_metadata_type(
                    name=metadata_type.name,
                    definition=metadata_type.definition,
                    concurrently=not allow_table_lock
                )
        return self.get_by_name(metadata_type.name)

    def can_update(self, metadata_type, allow_unsafe_updates=False):
        """
        Check if metadata type can be updated. Return bool,safe_changes,unsafe_changes

        Safe updates currently allow new search fields to be added, description to be changed.

        :param datacube.model.MetadataType metadata_type: updated MetadataType
        :param bool allow_unsafe_updates: Allow unsafe changes. Use with caution.
        :rtype: bool,list[change],list[change]
        """
        MetadataType.validate(metadata_type.definition)

        existing = self.get_by_name(metadata_type.name)
        if not existing:
            raise ValueError('Unknown metadata type %s, cannot update – '
                             'did you intend to add it?' % metadata_type.name)

        updates_allowed = {
            ('description',): changes.allow_any,
            # You can add new fields safely but not modify existing ones.
            ('dataset',): changes.allow_extension,
            ('dataset', 'search_fields'): changes.allow_extension
        }

        doc_changes = get_doc_changes(existing.definition, jsonify_document(metadata_type.definition))
        good_changes, bad_changes = changes.classify_changes(doc_changes, updates_allowed)

        for offset, old_val, new_val in good_changes:
            _LOG.info("Safe change in %s from %r to %r", _readable_offset(offset), old_val, new_val)

        for offset, old_val, new_val in bad_changes:
            _LOG.warning("Unsafe change in %s from %r to %r", _readable_offset(offset), old_val, new_val)

        return allow_unsafe_updates or not bad_changes, good_changes, bad_changes

    def update(self, metadata_type: MetadataType, allow_unsafe_updates=False, allow_table_lock=False):
        """
        Update a metadata type from the document. Unsafe changes will throw a ValueError by default.

        Safe updates currently allow new search fields to be added, description to be changed.

        :param datacube.model.MetadataType metadata_type: updated MetadataType
        :param bool allow_unsafe_updates: Allow unsafe changes. Use with caution.
        :param allow_table_lock:
            Allow an exclusive lock to be taken on the table while creating the indexes.
            This will halt other user's requests until completed.

            If false, creation will be slower and cannot be done in a transaction.
        :rtype: datacube.model.MetadataType
        """
        can_update, safe_changes, unsafe_changes = self.can_update(metadata_type, allow_unsafe_updates)

        if not safe_changes and not unsafe_changes:
            _LOG.warning("No changes detected for metadata type %s", metadata_type.name)
            return self.get_by_name(metadata_type.name)

        if not can_update:
            raise ValueError(f"Unsafe changes in {metadata_type.name}: " + (
                ", ".join(
                    _readable_offset(offset)
                    for offset, _, _ in unsafe_changes
                )
            ))

        _LOG.info("Updating metadata type %s", metadata_type.name)

        with self._db_connection() as connection:
            connection.update_metadata_type(
                name=metadata_type.name,
                definition=metadata_type.definition,
                concurrently=not allow_table_lock
            )

        self.get_by_name_unsafe.cache_clear()   # type: ignore[attr-defined]
        self.get_unsafe.cache_clear()           # type: ignore[attr-defined]
        return self.get_by_name(metadata_type.name)

    def update_document(self, definition, allow_unsafe_updates=False):
        """
        Update a metadata type from the document. Unsafe changes will throw a ValueError by default.

        Safe updates currently allow new search fields to be added, description to be changed.

        :param dict definition: Updated definition
        :param bool allow_unsafe_updates: Allow unsafe changes. Use with caution.
        :rtype: datacube.model.MetadataType
        """
        return self.update(self.from_doc(definition), allow_unsafe_updates=allow_unsafe_updates)

    # This is memoized in the constructor
    # pylint: disable=method-hidden
    def get_unsafe(self, id_):  # type: ignore
        with self._db_connection() as connection:
            record = connection.get_metadata_type(id_)
        if record is None:
            raise KeyError('%s is not a valid MetadataType id')
        return self._make_from_query_row(record)

    # This is memoized in the constructor
    # pylint: disable=method-hidden
    def get_by_name_unsafe(self, name):  # type: ignore
        with self._db_connection() as connection:
            record = connection.get_metadata_type_by_name(name)
        if not record:
            raise KeyError('%s is not a valid MetadataType name' % name)
        return self._make_from_query_row(record)

    def check_field_indexes(self, allow_table_lock=False,
                            rebuild_views=False, rebuild_indexes=False):
        """
        Create or replace per-field indexes and views.
        :param allow_table_lock:
            Allow an exclusive lock to be taken on the table while creating the indexes.
            This will halt other user's requests until completed.

            If false, creation will be slightly slower and cannot be done in a transaction.
        """
        with self._db_connection() as connection:
            connection.check_dynamic_fields(
                concurrently=not allow_table_lock,
                rebuild_indexes=rebuild_indexes,
                rebuild_views=rebuild_views,
            )

    def get_all(self):
        """
        Retrieve all Metadata Types

        :rtype: iter[datacube.model.MetadataType]
        """
        with self._db_connection() as connection:
            return self._make_many(connection.get_all_metadata_types())

    def _make_many(self, query_rows):
        """
        :rtype: list[datacube.model.MetadataType]
        """
        return (self._make_from_query_row(c) for c in query_rows)

    def _make_from_query_row(self, query_row):
        """
        :rtype: datacube.model.MetadataType
        """
        return self._make(query_row['definition'], query_row['id'])

    def _make(self, definition, id_=None):
        """
        :param dict definition:
        :param int id_:
        :rtype: datacube.model.MetadataType
        """
        return MetadataType(
            definition,
            dataset_search_fields=self._db.get_dataset_fields(definition),
            id_=id_
        )



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2020 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
"""
Module
"""
from pathlib import Path

import pytest

from datacube.testutils import write_files, assert_file_structure
from datacube.ui.common import get_metadata_path, _find_any_metadata_suffix, ui_path_doc_stream


def test_get_metadata_path():
    test_file_structure = {
        'directory_dataset': {
            'file1.txt': '',
            'file2.txt': '',
            'agdc-metadata.yaml.gz': ''
        },
        'file_dataset.tif': '',
        'file_dataset.tif.agdc-md.yaml': '',
        'dataset_metadata.yaml': '',
        'no_metadata.tif': '',
    }

    out_dir = write_files(test_file_structure)

    assert_file_structure(out_dir, test_file_structure)

    # A metadata file can be specified directly.
    path = get_metadata_path(out_dir.joinpath('dataset_metadata.yaml'))
    assert Path(path).absolute() == out_dir.joinpath('dataset_metadata.yaml').absolute()

    # A dataset directory will have an internal 'agdc-metadata' file.
    path = get_metadata_path(out_dir.joinpath('directory_dataset'))
    assert Path(path).absolute() == out_dir.joinpath('directory_dataset', 'agdc-metadata.yaml.gz').absolute()

    # Other out_dir can have a sibling file ending in 'agdc-md.yaml'
    path = get_metadata_path(out_dir.joinpath('file_dataset.tif'))
    assert Path(path).absolute() == out_dir.joinpath('file_dataset.tif.agdc-md.yaml').absolute()

    # URLs are always themselves
    example_url = 'http://localhost/dataset.yaml'
    url = get_metadata_path(example_url)
    assert url == example_url

    # Lack of metadata raises an error.
    with pytest.raises(ValueError):
        get_metadata_path(out_dir.joinpath('no_metadata.tif'))

    # Nonexistent dataset raises a ValueError.
    with pytest.raises(ValueError):
        get_metadata_path(out_dir.joinpath('missing-dataset.tif'))


def test_find_any_metatadata_suffix():
    files = write_files({
        'directory_dataset': {
            'file1.txt': '',
            'file2.txt': '',
            'agdc-metadata.json.gz': ''
        },
        'file_dataset.tif.agdc-md.yaml': '',
        'dataset_metadata.YAML': '',
        'no_metadata.tif': '',
        'ambigous.yml': '',
        'ambigous.yaml': '',
    })

    path = _find_any_metadata_suffix(files.joinpath('dataset_metadata'))
    assert Path(path).absolute() == files.joinpath('dataset_metadata.YAML').absolute()

    path = _find_any_metadata_suffix(files.joinpath('directory_dataset', 'agdc-metadata'))
    assert Path(path).absolute() == files.joinpath('directory_dataset', 'agdc-metadata.json.gz').absolute()

    path = _find_any_metadata_suffix(files.joinpath('file_dataset.tif.agdc-md'))
    assert Path(path).absolute() == files.joinpath('file_dataset.tif.agdc-md.yaml').absolute()

    # Returns none if none exist
    path = _find_any_metadata_suffix(files.joinpath('no_metadata'))
    assert path is None

    with pytest.raises(ValueError):
        _find_any_metadata_suffix(files.joinpath('ambigous'))


def test_ui_path_doc_stream(httpserver):
    filename = 'dataset_metadata.yaml'
    file_content = ''
    out_dir = write_files({filename: file_content})

    httpserver.expect_request(filename).respond_with_data(file_content)

    input_paths = [Path(out_dir) / 'dataset_metadata.yaml', httpserver.url_for(filename)]

    for input_path, (doc, resolved_path) in zip(input_paths, ui_path_doc_stream(input_paths)):
        assert doc == {}
        assert input_path == resolved_path



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2020 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
from setuptools import setup, find_packages

setup(
    name='dc_tests_io',
    version="1.0",
    description='Example "bad drivers" for testing driver loading protections',
    author='AGDC Collaboration',
    packages=find_packages(),

    entry_points={
        'datacube.plugins.io.read': [
            'bad_end_point=dc_tests_io.nosuch.module:init',
            'failing_end_point_throw=dc_tests_io.dummy:fail_to_init',
            'failing_end_point_none=dc_tests_io.dummy:init_to_none',
        ]
    }
)



#!/usr/bin/env python
# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2020 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
"""
Query datasets.
"""

import csv
import datetime
import shutil
import sys
from functools import partial

import click
from dateutil import tz
from psycopg2._range import Range
from functools import singledispatch

from datacube.ui import click as ui
from datacube.ui.click import CLICK_SETTINGS

PASS_INDEX = ui.pass_index('datacube-search')


def printable_values(d):
    return {k: printable(v) for k, v in d.items()}


def write_pretty(out_f, field_names, search_results, terminal_size=shutil.get_terminal_size()):
    """
    Output in a human-readable text format. Inspired by psql's expanded output.
    """
    terminal_width = terminal_size[0]
    record_num = 1

    field_header_width = max(len(name) for name in field_names)
    field_output_format = '{:<' + str(field_header_width) + '} | {}'

    for result in search_results:
        separator_line = '-[ {} ]'.format(record_num)
        separator_line += '-' * (terminal_width - len(separator_line) - 1)
        click.echo(separator_line, file=out_f)

        for name, value in sorted(result.items()):
            click.echo(
                field_output_format.format(name, printable(value)),
                file=out_f
            )

        record_num += 1


def write_csv(out_f, field_names, search_results):
    """
    Output as a CSV.
    """
    search_results = list(search_results)
    writer = csv.DictWriter(out_f, tuple(sorted(field_names)))
    writer.writeheader()
    writer.writerows(
        (
            printable_values(d) for d in
            search_results
        )
    )


OUTPUT_FORMATS = {
    'csv': write_csv,
    'pretty': write_pretty
}


@click.group(help="Search the Data Cube", context_settings=CLICK_SETTINGS)
@ui.global_cli_options
@click.option('-f',
              type=click.Choice(list(OUTPUT_FORMATS)),
              default='pretty', show_default=True,
              help='Output format')
@click.pass_context
def cli(ctx, f):
    ctx.obj['write_results'] = partial(OUTPUT_FORMATS[f], sys.stdout)


@cli.command()
@ui.parsed_search_expressions
@PASS_INDEX
@click.pass_context
def datasets(ctx, index, expressions):
    """
    Search available Datasets
    """
    ctx.obj['write_results'](
        sorted(index.datasets.get_field_names()),
        index.datasets.search_summaries(**expressions)
    )


@cli.command('product-counts')
@click.argument('period', nargs=1)
@ui.parsed_search_expressions
@PASS_INDEX
def product_counts(index, period, expressions):
    """
    Count product Datasets available by period

    PERIOD: eg. 1 month, 6 months, 1 year
    """
    for product, series in index.datasets.count_by_product_through_time(period, **expressions):
        click.echo(product.name)
        for timerange, count in series:
            formatted_dt = _assume_utc(timerange[0]).strftime("%Y-%m-%d")
            click.echo('    {}: {}'.format(formatted_dt, count))


@singledispatch
def printable(val):
    return val


@printable.register(type(None))
def printable_none(val):
    return ''


@printable.register(datetime.datetime)
def printable_dt(val):
    """
    :type val: datetime.datetime
    """
    return _assume_utc(val).isoformat()


def _assume_utc(val):
    if val.tzinfo is None:
        return val.replace(tzinfo=tz.tzutc())
    else:
        return val.astimezone(tz.tzutc())


@printable.register(Range)
def printable_r(val):
    """
    :type val: psycopg2._range.Range
    """
    if val.lower_inf:
        return printable(val.upper)
    if val.upper_inf:
        return printable(val.lower)

    return '{} to {}'.format(printable(val.lower), printable(val.upper))


if __name__ == '__main__':
    cli()



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2020 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
import logging
from copy import deepcopy
from pathlib import Path
from types import SimpleNamespace

import numpy
import pytest
import yaml

try:
    from yaml import CSafeDumper as SafeDumper  # type: ignore
    from yaml import CSafeLoader as SafeLoader  # type: ignore
except ImportError:
    from yaml import SafeLoader, SafeDumper  # type: ignore

import rasterio
import xarray as xr
from affine import Affine
from datacube.api.core import Datacube
from datacube.utils import geometry

pytest.importorskip("dcio_example.xarray_3d")  # skip this test if 3d driver is not installed
_LOG = logging.getLogger(__name__)
_LOG.setLevel(logging.DEBUG)

PROJECT_ROOT = Path(__file__).parents[1]
GEDI_TEST_DATA = PROJECT_ROOT / "tests" / "data" / "lbg" / "gedi"

GEDI_PRODUCT = SimpleNamespace(
    dataset_types=GEDI_TEST_DATA / "GEDI02_B_3d_format.yaml",
    crs="EPSG:4326",
)
# Params common to 2D and 3D GEDI products

GEDI_PRODUCTS = {
    "2D": SimpleNamespace(
        name="gedi_l2b",
        measurements=["cover", "pai"],
        wavelengths=None,  # None for 2D products
        index_yaml=GEDI_TEST_DATA / "{product_id}" / "__{product_id}_gedi_l2b.yaml",
    ),
    "3D": SimpleNamespace(  # 3D with wavelength range
        name="gedi_l2b_{measurement}",
        measurements=["cover_z", "pai_z"],
        wavelengths=(20, 30),  # (5, 150),  # Heights in GEDI `*_z` products
        index_yaml=GEDI_TEST_DATA
        / "{product_id}"
        / "__{product_id}_gedi_l2b_{measurement}.yaml",
    ),
    "3D2": SimpleNamespace(  # 3D with single wavelength value
        name="gedi_l2b_{measurement}",
        measurements=["cover_z", "pai_z"],
        wavelengths=40,  # (5, 150),  # Heights in GEDI `*_z` products
        index_yaml=GEDI_TEST_DATA
        / "{product_id}"
        / "__{product_id}_gedi_l2b_{measurement}.yaml",
    ),
}

GEDI_PRODUCT_IDS = [
    SimpleNamespace(
        pid="GEDI02_B_2019228092851_O03828_T04236_02_001_01",
        size=(548, 415),
        affine=Affine(
            0.00027778,
            0.0,
            149.04036673536626,
            0.0,
            -0.00027778,
            -35.30377268593081,
        ),
        time=("2019-08", "2019-08"),
    ),
    SimpleNamespace(
        pid="GEDI02_B_2019294155401_O04856_T03859_02_001_01",
        size=(551, 420),
        affine=Affine(
            0.00027778,
            0.0,
            149.03966950632497,
            0.0,
            -0.00027778,
            -35.30265746130061,
        ),
        time=("2019-10", "2019-10"),
    ),
]
# Product IDs must be in chronological order


def custom_dumb_fuser(dst, src):
    dst[:] = src[:]


ignore_me = pytest.mark.xfail(
    True, reason="get_data/get_description still to be fixed in Unification"
)


@pytest.fixture(scope="module")
def original_data():
    """Load and cache the original NetCDF data from disk.

    This is the same data that gets indexed, hence the files have the suffix
    `.xarray_3d` because they are 3D NetCDF files which need the `xarray_3d` ODC
    driver (as opposed to the normal NetCDF driver). All available files get
    loaded once per session and returned as a dict of `{product_id_measurement:
    data_array}`.
    """
    data = {}
    for path in GEDI_TEST_DATA.rglob("*.xarray_3d"):
        datum = xr.open_dataset(path, engine="netcdf4")["array"]
        # Rename x/y to lat/lon because of the crs used
        # Index as file name without .xarray_3d extension
        data[path.name[:-10]] = datum.rename({"x": "longitude", "y": "latitude"})
    yield data


@pytest.fixture(scope="function", params=["3D", "3D2", "2D"])
def product_def(request):
    """GEDI 3D and 2D product parameters."""
    yield GEDI_PRODUCTS[request.param]


@pytest.fixture
def invalid_dataset_type_paths(tmpdir):
    """Prepare a series of invalid dataset type yamls.

    Returns: a dict of {<expected error message>: <yaml path>} for each invalid
    yaml.
    """
    with GEDI_PRODUCT.dataset_types.open() as fh:
        documents = [
            doc
            for doc in yaml.load_all(fh, Loader=SafeLoader)
            if doc["name"] == "gedi_l2b_cover_z"
        ]
    assert len(documents) == 1, "Test cannot alter product definition"
    invalid_docs = {}

    # Error: Missing extra_dimensions
    doc = deepcopy(documents[0])
    del doc["extra_dimensions"]  # Dropped item
    invalid_docs["extra_dimensions is not defined"] = doc

    # Error: extra_dimensions.name mismatch extra_dim.dimension
    doc = deepcopy(documents[0])
    doc["extra_dimensions"][0]["name"] = "invalid"  # Spurious name
    invalid_docs["Dimension z is not defined in extra_dimension"] = doc

    # Error: extra_dimensions.values length mismatch extra_dim.spectral_definition_map
    doc = deepcopy(documents[0])
    doc["measurements"][0]["spectral_definition"] = [
        {
            "wavelength": [w, w + 1],
            "response": [w / 10, (w + 1) / 10],
        }
        for w in doc["extra_dimensions"][0]["values"]
        + [-9999]  # Spurious trailing item
    ]
    invalid_docs[
        "spectral_definition should be the same length as values for extra_dim z"
    ] = doc

    # Error: mismatching spectral wavelength and response
    doc = deepcopy(documents[0])
    doc["measurements"][0]["spectral_definition"] = [
        {
            "wavelength": [w, w + 1, w + 2],  # Wavelength longer than response
            "response": [w / 10, (w + 1) / 10],
        }
        for w in doc["extra_dimensions"][0]["values"]
    ]
    invalid_docs[
        "spectral_definition_map: wavelength should be the same length as response"
    ] = doc

    invalid_paths = {}
    for name, invalid_doc in invalid_docs.items():
        path = Path(tmpdir) / (
            name.lower().replace(" ", "_").replace(".", "_") + ".yaml"
        )
        with path.open("w") as fh:
            yaml.dump(invalid_doc, fh, Dumper=SafeDumper)
        invalid_paths[name] = path
    yield invalid_paths


@pytest.fixture
def product_with_spectral_map(tmpdir):
    """Create a copy of input yaml with a spectral map.

    Returns the path string to the copy.
    """
    with GEDI_PRODUCT.dataset_types.open() as fh:
        documents = [
            doc
            for doc in yaml.load_all(fh, Loader=SafeLoader)
            if doc["name"] == "gedi_l2b_cover_z"
        ]
    assert len(documents) == 1, "Test cannot alter product definition"
    doc = deepcopy(documents[0])
    doc["measurements"][0]["spectral_definition"] = [
        {
            "wavelength": [w, w + 1],
            "response": [w / 10, (w + 1) / 10],
        }
        for w in doc["extra_dimensions"][0]["values"]
    ]
    path = Path(tmpdir) / GEDI_PRODUCT.dataset_types.name
    with path.open("w") as fh:
        yaml.dump(doc, fh, Dumper=SafeDumper)
    return path


@pytest.fixture(
    scope="function",
    params=["Without_spectral_map", "With_spectral_map"],
)
def dataset_types(product_with_spectral_map, request):
    """GEDI datasets types with/out spectral_map."""
    if request.param.startswith("Without"):
        yield GEDI_PRODUCT.dataset_types
    else:
        yield product_with_spectral_map


@pytest.mark.usefixtures("default_metadata_type")
def test_missing_extra_dimensions(clirunner, invalid_dataset_type_paths):
    """Test error on invalid product definition."""
    for expected_msg, path in invalid_dataset_type_paths.items():
        with pytest.raises(ValueError) as exc_info:
            clirunner(["-v", "product", "add", str(path)])
        assert str(exc_info.value).startswith(expected_msg)


@pytest.mark.usefixtures("default_metadata_type")
def test_indexing(clirunner, index, product_def):
    """Test indexing features for 2D and 3D products.

    A few no-op indexing commands are tested as well as a simple load with shape
    check only.
    """
    product_id = GEDI_PRODUCT_IDS[0]
    measurement = product_def.measurements[0]
    index_yaml = str(product_def.index_yaml).format(
        product_id=product_id.pid,
        measurement=measurement,
    )

    # Add the GEDI Dataset Types
    clirunner(["-v", "product", "add", str(GEDI_PRODUCT.dataset_types)])

    # Index the Datasets
    #  - do test run first to increase test coverage
    clirunner(["-v", "dataset", "add", "--dry-run", str(index_yaml)])

    #  - do actual indexing
    clirunner(["-v", "dataset", "add", str(index_yaml)])

    #  - this will be no-op but with ignore lineage
    clirunner(
        [
            "-v",
            "dataset",
            "add",
            "--confirm-ignore-lineage",
            str(index_yaml),
        ]
    )

    # Test no-op update
    for policy in ["archive", "forget", "keep"]:
        clirunner(
            [
                "-v",
                "dataset",
                "update",
                "--dry-run",
                "--location-policy",
                policy,
                str(index_yaml),
            ]
        )

        # Test no changes needed update
        clirunner(
            [
                "-v",
                "dataset",
                "update",
                "--location-policy",
                policy,
                str(index_yaml),
            ]
        )

    dc = Datacube(index=index)
    check_open_with_dc_simple(dc, product_def, [product_id], measurement)


@pytest.mark.usefixtures("default_metadata_type")
def test_indexing_with_spectral_map(clirunner, index, dataset_types):
    """Test indexing features with spectral map."""
    product_id = GEDI_PRODUCT_IDS[0]
    product_def = GEDI_PRODUCTS["3D"]
    measurement = product_def.measurements[0]
    index_yaml = str(product_def.index_yaml).format(
        product_id=product_id.pid,
        measurement=measurement,
    )

    # Add the GEDI Dataset Types
    clirunner(["-v", "product", "add", str(dataset_types)])

    # Index the Dataset
    clirunner(["-v", "dataset", "add", str(index_yaml)])
    dc = Datacube(index=index)
    check_open_with_dc_simple(dc, product_def, [product_id], measurement)


@pytest.mark.usefixtures("default_metadata_type")
def test_end_to_end_multitime(clirunner, index, product_def, original_data):
    """Test simple indexing but for multiple measurements and wavelengths."""
    dc = Datacube(index=index)

    # Add the GEDI Dataset Types
    clirunner(["-v", "product", "add", str(GEDI_PRODUCT.dataset_types)])

    for idx, measurement in enumerate(product_def.measurements):
        for product_id in GEDI_PRODUCT_IDS:
            index_yaml = str(product_def.index_yaml).format(
                product_id=product_id.pid,
                measurement=measurement,
            )
            # Index the Datasets
            clirunner(["-v", "dataset", "add", str(index_yaml)])

        if idx == 0:  # Full check for the first measurement only
            # Check data for all product IDs
            check_open_with_dc_contents(
                dc, product_def, GEDI_PRODUCT_IDS, measurement, original_data
            )
            # check_open_with_grid_workflow(index)
            # Only test first product ID with dss
            check_load_via_dss(
                dc, product_def, GEDI_PRODUCT_IDS[:1], measurement, original_data
            )
        else:
            check_open_with_dc_simple(dc, product_def, GEDI_PRODUCT_IDS, measurement)


def check_loaded_vs_original(data, orig, product_def):
    """Check that the `data` against the original data from disk."""
    data_t = data.isel(time=0)  # Only 1 time slice for now
    if product_def.wavelengths:
        # 3D: Compare values for all loaded wavelengths
        wavelengths = (
            (product_def.wavelengths,) * 2
            if isinstance(product_def.wavelengths, int)
            else product_def.wavelengths
        )
        for wavelength in range(wavelengths[0], wavelengths[1] + 1, 5):
            data_w = data_t.sel(z=wavelength)
            orig_w = orig.sel(z=wavelength)
            assert numpy.array_equal(data_w.values, orig_w.values)
    else:
        # 2D: Compare values
        # TODO: When the bug preventing geobox for 2D is fixed, uncomment the
        # following line
        # assert numpy.array_equal(data_t.values, orig.values)
        _LOG.info(f"\n{str(numpy.array_equal(data_t.values, orig.values)):~^80}\n")


def load_with_dc(
    dc, product_def, product_id, measurement, time=None, datasets=None, dask_chunks=None
):
    """Load data with dc, with settable params.

    If `datasets` is specified, the no `product` is used in the load command.
    `dask_chunks` get passed as-is.
    """
    params = SimpleNamespace(
        measurements=[measurement],
        like=geometry.GeoBox(
            *product_id.size,
            product_id.affine,
            GEDI_PRODUCT.crs,
        ),
        dask_chunks=dask_chunks,
    )
    if time:
        params.time = time
    if product_def.wavelengths:
        params.z = product_def.wavelengths

    if datasets:
        params.datasets = datasets
    else:
        params.product = product_def.name.format(measurement=measurement)
    _LOG.info(f"DC Loading {params}")
    data = dc.load(**params.__dict__)
    _LOG.info(f"DC Loaded\n{data}\n{'-'*80}")
    return data


def check_open_with_dc_simple(dc, product_def, product_ids, measurement):
    """Check data can be loaded and has the right shape.

    Only the first of `product_ids` is tested. The actual contents of the loaded
    data are not checked, only their shape.
    """
    product_id = product_ids[0]
    data = load_with_dc(dc, product_def, product_id, measurement)
    expected = [len(product_ids), *product_id.size[::-1]]
    if product_def.wavelengths:
        wlen = (
            1
            if isinstance(product_def.wavelengths, int)
            else len(range(*product_def.wavelengths, 5)) + 1
        )
        expected.insert(1, wlen)
    assert list(data[measurement].shape) == expected


def check_open_with_dc_contents(
    dc, product_def, product_ids, measurement, original_data
):
    """Check simple and dask loads.

    Simple load is checked against the original file, loaded from disk. Lazy
    loading with `dask_chunks` is then compared to the simple load.
    """
    for product_id in product_ids:
        orig = original_data[f"{product_id.pid}_{measurement}"]
        data = load_with_dc(
            dc, product_def, product_id, measurement, time=product_id.time
        )
        check_loaded_vs_original(data[measurement], orig, product_def)

    # Use the last data from the for loop above, i.e., time slice is the last
    # one, assuming product_ids are in chronological order
    data_array = data[measurement]

    # Simpler checks, but using dask
    with rasterio.Env():
        dask_chunks = SimpleNamespace(time=1, latitude=200, longitude=200)
        # We will compare data at the intersection of 4 chunks
        target = SimpleNamespace(
            time=0, latitude=slice(150, 250), longitude=slice(150, 250)
        )
        if product_def.wavelengths:
            dask_chunks.z = 1
            target.z = 0
        lazy_data_array = load_with_dc(
            dc,
            product_def,
            product_id,
            measurement,
            time=product_id.time,
            dask_chunks=dask_chunks.__dict__,
        )[measurement]
        assert lazy_data_array.data.dask
        assert lazy_data_array.ndim == data_array.ndim
        # Checking the target area is not full of nodata
        assert dataarray_has_valid_data(data_array.isel(target.__dict__))
        assert lazy_data_array.isel(target.__dict__).equals(
            data_array.isel(target.__dict__)
        )


def dataarray_has_valid_data(da):
    """Check xarray has valid data."""
    return da.size and not (da.values == da.nodata).all()


def check_open_with_grid_workflow(index):
    """Not implemented"""
    pass


def check_load_via_dss(dc, product_def, product_ids, measurement, original_data):
    """Check dataset can be searched and loaded, and has the right shape.

    Only the first of `product_ids` is tested. The actual contents of the loaded
    data are not checked, only their shape.
    """
    product_id = product_ids[0]
    product = product_def.name.format(measurement=measurement)
    datasets = dc.find_datasets(product=product, time=product_id.time)
    assert len(datasets) > 0
    data = load_with_dc(
        dc,
        product_def,
        product_id,
        measurement,
        datasets=datasets,
    )
    expected = [len(product_ids), *product_id.size[::-1]]
    if product_def.wavelengths:
        wlen = (
            1
            if isinstance(product_def.wavelengths, int)
            else len(range(*product_def.wavelengths, 5)) + 1
        )
        expected.insert(1, wlen)
    assert list(data[measurement].shape) == expected



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2020 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
"""
The start of some general purpose utilities for generating test data.

At the moment it's not very generic, but can continue to be extended in that fashion.

Hypothesis is used for most of the data generation, which will hopefully improve the rigour
of our tests when we can roll it into more tests.
"""
import string
from datetime import datetime, timedelta
from pathlib import Path

import numpy as np
import rasterio
import yaml
from hypothesis.strategies import sampled_from, datetimes, composite, floats, lists, text, uuids

from datacube.utils.geometry import CRS, point

DATE_FORMAT = '%Y-%m-%d %H:%M:%S'

# Bounds retrieved from http://www.epsg-registry.org
# Code, lat_bounds, lon_bounds
EPSG_CODES = [('EPSG:32755', (-80, 0), (144, 150))]
crses = sampled_from(EPSG_CODES)
REASONABLE_DATES = datetimes(min_value=datetime(1980, 1, 1), max_value=datetime(2030, 1, 1))
FILENAMES = text(alphabet=string.ascii_letters + string.digits + '-_', min_size=3)


@composite
def bounds(draw, bounds):
    return draw(
        lists(
            floats(*bounds), min_size=2, max_size=2
        ).map(sorted).filter(lambda x: x[0] < x[1]))


@composite
def bboxes(draw, lat_bounds=(-90, 90), lon_bounds=(-180, 180)):
    """
    https://tools.ietf.org/html/rfc7946#section-5
    """
    # Use 3 dim positions even if we only need 2
    min_lat, max_lat = draw(bounds(lat_bounds))
    min_lon, max_lon = draw(bounds(lon_bounds))

    return min_lon, min_lat, max_lon, max_lat


@composite
def extents(draw, lat_bounds, lon_bounds):
    center_dt = draw(REASONABLE_DATES)

    time_range = timedelta(seconds=12)
    min_lon, min_lat, max_lon, max_lat = draw(bboxes(lat_bounds, lon_bounds))
    return {
        'center_dt': center_dt.strftime(DATE_FORMAT),
        'from_dt': (center_dt - time_range).strftime(DATE_FORMAT),
        'to_dt': (center_dt + time_range).strftime(DATE_FORMAT),
        'coord': {
            'll': {'lat': min_lat, 'lon': min_lon},
            'lr': {'lat': min_lat, 'lon': max_lon},
            'ul': {'lat': max_lat, 'lon': min_lon},
            'ur': {'lat': max_lat, 'lon': max_lon},
        }
    }


def _extent_point_projector(crs):
    crs = CRS(crs)

    def reproject_point(pos):
        pos = point(pos['lon'], pos['lat'], CRS('EPSG:4326'))
        coords = pos.to_crs(crs).coords[0]
        return {'x': coords[0], 'y': coords[1]}

    return reproject_point


def extent_to_grid_spatial(extent, crs):
    """Convert an extent in WGS84 to a grid spatial in the supplied CRS"""
    reprojector = _extent_point_projector(crs)
    return {
        'projection': {
            'geo_ref_points': {
                corner: reprojector(pos)
                for corner, pos in extent['coord'].items()
            },
            'spatial_reference': crs
        }
    }


@composite
def acquisition_details(draw):
    return {
        'aos': draw(REASONABLE_DATES),
        'groundstation': {
            'code': draw(text(alphabet=string.ascii_letters, min_size=3, max_size=5))
        },
        'los': draw(REASONABLE_DATES)
    }


@composite
def scene_datasets(draw):
    """
    Generate random test Landsat 5 Scene Datasets
    """
    crs, lat_bounds, lon_bounds = draw(crses)
    extent = draw(extents(lat_bounds, lon_bounds))
    return {
        'id': str(draw(uuids())),
        'acquisition': draw(acquisition_details()),
        'creation_dt': draw(REASONABLE_DATES),
        'extent': extent,
        'grid_spatial': extent_to_grid_spatial(extent, crs),
        'image': {
            'bands': {
                '1': {
                    'path': draw(FILENAMES) + '.tif'
                }
            }
        },
        'format': {
            'name': 'GeoTIFF'},
        'instrument': {
            'name': 'TM'},
        'lineage': {
            'source_datasets': {}},
        'platform': {
            'code': 'LANDSAT_5'},
        'processing_level': 'P54',
        'product_type': 'nbar'
    }


def write_test_scene_to_disk(dataset_dict, tmpdir):
    tmpdir = Path(str(tmpdir))
    # Make directory name
    dir_name = dataset_dict['platform']['code'] + dataset_dict['id']

    # Create directory
    new_dir = tmpdir / dir_name
    new_dir.mkdir(exist_ok=True)

    _make_geotiffs(new_dir, dataset_dict)
    dataset_file = new_dir / 'agdc-metadata.yaml'
    with dataset_file.open('w') as out:
        yaml.safe_dump(dataset_dict, out)
    return dataset_file


def _make_geotiffs(output_dir, dataset_dict, shape=(100, 100)):
    """
    Generate test GeoTIFF files into ``output_dir``, one per band, from a dataset dictionary

    :return: a dictionary mapping band_number to filename, eg::

        {
            0: '/tmp/tiffs/band01_time01.tif',
            1: '/tmp/tiffs/band02_time01.tif'
        }
    """
    tiffs = {}
    width, height = shape
    pixel_width = (dataset_dict['grid_spatial']['projection']['geo_ref_points']['ul']['x'] -
                   dataset_dict['grid_spatial']['projection']['geo_ref_points']['ur']['x']) / width
    pixel_height = (dataset_dict['grid_spatial']['projection']['geo_ref_points']['ul']['y'] -
                    dataset_dict['grid_spatial']['projection']['geo_ref_points']['ll']['y']) / height
    metadata = {'count': 1,
                'crs': dataset_dict['grid_spatial']['projection']['spatial_reference'],
                'driver': 'GTiff',
                'dtype': 'int16',
                'width': width,
                'height': height,
                'nodata': -999.0,
                'transform': [pixel_width,
                              0.0,
                              dataset_dict['grid_spatial']['projection']['geo_ref_points']['ul']['x'],
                              0.0,
                              pixel_height,
                              dataset_dict['grid_spatial']['projection']['geo_ref_points']['ul']['y']]}

    for band_num, band_info in dataset_dict['image']['bands'].items():
        path = Path(output_dir) / band_info['path']
        with rasterio.open(path, 'w', **metadata) as dst:
            # Write data in "corners" (rounded down by 100, for a size of 100x100)
            data = np.zeros((height, width), dtype=np.int16)
            data[:] = np.arange(height * width
                                ).reshape((height, width)) + 10 * int(band_num)
            dst.write(data, 1)
        tiffs[band_num] = path
    return tiffs



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2020 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
import logging
from contextlib import contextmanager

from datacube.drivers.postgres import PostgresDb, PostgresDbAPI
from datacube.index.postgres._transaction import PostgresTransaction
from datacube.index.postgres._datasets import DatasetResource  # type: ignore
from datacube.index.postgres._metadata_types import MetadataTypeResource
from datacube.index.postgres._products import ProductResource
from datacube.index.postgres._users import UserResource
from datacube.index.abstract import AbstractIndex, AbstractIndexDriver, default_metadata_type_docs, AbstractTransaction
from datacube.model import MetadataType
from datacube.utils.geometry import CRS

_LOG = logging.getLogger(__name__)


class Index(AbstractIndex):
    """
    Access to the datacube index.

    DON'T INITIALISE THIS DIRECTLY (it will break in the future). Use `datacube.index.index_connect()` or
    access property ``.index`` on your existing :class:`datacube.api.core.Datacube`.

    These are thread safe. But not multiprocess safe once a connection is made (db connections cannot be shared
    between processes) You can close idle connections before forking by calling close(), provided you know no
    other connections are active. Or else use a separate instance of this class in each process.

    :ivar datacube.index._datasets.DatasetResource datasets: store and retrieve :class:`datacube.model.Dataset`
    :ivar datacube.index._products.ProductResource products: store and retrieve :class:`datacube.model.DatasetType`\
    (should really be called Product)
    :ivar datacube.index._metadata_types.MetadataTypeResource metadata_types: store and retrieve \
    :class:`datacube.model.MetadataType`
    :ivar UserResource users: user management

    :type users: datacube.index._users.UserResource
    :type datasets: datacube.index._datasets.DatasetResource
    :type products: datacube.index._products.ProductResource
    :type metadata_types: datacube.index._metadata_types.MetadataTypeResource
    """

    supports_transactions = True

    def __init__(self, db: PostgresDb) -> None:
        self._db = db

        self._users = UserResource(db, self)
        self._metadata_types = MetadataTypeResource(db, self)
        self._products = ProductResource(db, self)
        self._datasets = DatasetResource(db, self)

    @property
    def users(self) -> UserResource:
        return self._users

    @property
    def metadata_types(self) -> MetadataTypeResource:
        return self._metadata_types

    @property
    def products(self) -> ProductResource:
        return self._products

    @property
    def datasets(self) -> DatasetResource:
        return self._datasets

    @property
    def url(self) -> str:
        return str(self._db.url)

    @classmethod
    def from_config(cls, config, application_name=None, validate_connection=True):
        db = PostgresDb.from_config(config, application_name=application_name,
                                    validate_connection=validate_connection)
        return cls(db)

    @classmethod
    def get_dataset_fields(cls, doc):
        return PostgresDb.get_dataset_fields(doc)

    def init_db(self, with_default_types=True, with_permissions=True):
        is_new = self._db.init(with_permissions=with_permissions)

        if is_new and with_default_types:
            _LOG.info('Adding default metadata types.')
            for doc in default_metadata_type_docs():
                self.metadata_types.add(self.metadata_types.from_doc(doc), allow_table_lock=True)

        return is_new

    def close(self):
        """
        Close any idle connections database connections.

        This is good practice if you are keeping the Index instance in scope
        but wont be using it for a while.

        (Connections are normally closed automatically when this object is deleted: ie. no references exist)
        """
        self._db.close()

    @property
    def index_id(self) -> str:
        return f"legacy_{self.url}"

    def transaction(self) -> AbstractTransaction:
        return PostgresTransaction(self._db, self.index_id)

    def create_spatial_index(self, crs: CRS) -> None:
        _LOG.warning("postgres driver does not support spatio-temporal indexes")

    def __repr__(self):
        return "Index<db={!r}>".format(self._db)

    @contextmanager
    def _active_connection(self, transaction: bool = False) -> PostgresDbAPI:
        """
        Context manager representing a database connection.

        If there is an active transaction for this index in the current thread, the connection object from that
        transaction is returned, with the active transaction remaining in control of commit and rollback.

        If there is no active transaction and the transaction argument is True, a new transactionised connection
        is returned, with this context manager handling commit and rollback.

        If there is no active transaction and the transaction argument is False (the default), a new connection
        is returned with autocommit semantics.

        Note that autocommit behaviour is NOT available if there is an active transaction for the index
        and the active thread.

        :param transaction: Use a transaction if one is not already active for the thread.
        :return: A PostgresDbAPI object, with the specified transaction semantics.
        """
        trans = self.thread_transaction()
        closing = False
        if trans is not None:
            # Use active transaction
            yield trans._connection
        elif transaction:
            closing = True
            with self._db._connect() as conn:
                conn.begin()
                try:
                    yield conn
                    conn.commit()
                except Exception:  # pylint: disable=broad-except
                    conn.rollback()
                    raise
        else:
            closing = True
            # Autocommit behaviour:
            with self._db._connect() as conn:
                yield conn


class DefaultIndexDriver(AbstractIndexDriver):
    aliases = ['postgres']

    @staticmethod
    def connect_to_index(config, application_name=None, validate_connection=True):
        return Index.from_config(config, application_name, validate_connection)

    @staticmethod
    def metadata_type_from_doc(definition: dict) -> MetadataType:
        """
        :param definition:
        """
        MetadataType.validate(definition)  # type: ignore
        return MetadataType(definition,
                            dataset_search_fields=Index.get_dataset_fields(definition))


def index_driver_init():
    return DefaultIndexDriver()



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2020 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
import numpy as np
import collections.abc
from types import SimpleNamespace
from typing import Tuple
from affine import Affine

# This is numeric code, short names make sense in this context, so disabling
# "invalid name" checks for the whole file
# pylint: disable=invalid-name


class WindowFromSlice:
    """ Translate numpy slices to rasterio window tuples.
    """
    def __getitem__(self, roi):
        if roi is None:
            return None

        if not isinstance(roi, collections.abc.Sequence) or len(roi) != 2:
            raise ValueError("Need 2d roi")

        row, col = roi
        return ((0 if row.start is None else row.start, row.stop),
                (0 if col.start is None else col.start, col.stop))


w_ = WindowFromSlice()


def polygon_path(x, y=None):
    """A little bit like numpy.meshgrid, except returns only boundary values and
    limited to 2d case only.

    Examples:
      [0,1], [3,4] =>
      array([[0, 1, 1, 0, 0],
             [3, 3, 4, 4, 3]])

      [0,1] =>
      array([[0, 1, 1, 0, 0],
             [0, 0, 1, 1, 0]])
    """

    if y is None:
        y = x

    return np.vstack([
        np.vstack([x, np.full_like(x, y[0])]).T,
        np.vstack([np.full_like(y, x[-1]), y]).T[1:],
        np.vstack([x, np.full_like(x, y[-1])]).T[::-1][1:],
        np.vstack([np.full_like(y, x[0]), y]).T[::-1][1:]]).T


def gbox_boundary(gbox, pts_per_side=16):
    """Return points in pixel space along the perimeter of a GeoBox, or a 2d array.

    """
    H, W = gbox.shape[:2]
    xx = np.linspace(0, W, pts_per_side, dtype='float32')
    yy = np.linspace(0, H, pts_per_side, dtype='float32')

    return polygon_path(xx, yy).T[:-1]


def roi_boundary(roi, pts_per_side=2):
    """
    Get boundary points from a 2d roi.

    roi needs to be in the normalised form, i.e. no open-ended start/stop, see roi_normalise

    :returns: Nx2 float32 array of X,Y points on the perimeter of the envelope defined by `roi`
    """
    yy, xx = roi
    xx = np.linspace(xx.start, xx.stop, pts_per_side, dtype='float32')
    yy = np.linspace(yy.start, yy.stop, pts_per_side, dtype='float32')

    return polygon_path(xx, yy).T[:-1]


def align_down(x, align):
    return x - (x % align)


def align_up(x, align):
    return align_down(x+(align-1), align)


def scaled_down_roi(roi, scale: int):
    return tuple(slice(s.start//scale,
                       align_up(s.stop, scale)//scale) for s in roi)


def scaled_up_roi(roi, scale: int, shape=None):
    roi = tuple(slice(s.start*scale,
                      s.stop*scale) for s in roi)
    if shape is not None:
        roi = tuple(slice(min(dim, s.start),
                          min(dim, s.stop))
                    for s, dim in zip(roi, shape))
    return roi


def scaled_down_shape(shape, scale: int):
    return tuple(align_up(s, scale)//scale for s in shape)


def roi_shape(roi):
    def slice_dim(s):
        return s.stop if s.start is None else s.stop - s.start

    if isinstance(roi, slice):
        roi = (roi,)

    return tuple(slice_dim(s) for s in roi)


def roi_is_empty(roi):
    return any(d <= 0 for d in roi_shape(roi))


def roi_is_full(roi, shape):
    """
    Check if ROI covers the entire region.

    :returns: True if roi covers region from (0,..) -> shape
              False otherwise
    """
    def slice_full(s, n):
        return s.start in (0, None) and s.stop in (n, None)

    if isinstance(roi, slice):
        roi = (roi,)
        shape = (shape,)

    return all(slice_full(s, n) for s, n in zip(roi, shape))


def roi_normalise(roi, shape):
    """
    Fill in missing .start/.stop, also deal with negative values, which are
    treated as offsets from the end.

    .step parameter is left unchanged.

    Example:
          np.s_[:3, 4:  ], (10, 20) -> np._s[0:3, 4:20]
          np.s_[:3,  :-3], (10, 20) -> np._s[0:3, 0:17]

    """

    def fill_if_none(x, val_if_none):
        return val_if_none if x is None else x

    def norm_slice(s, n):
        start = fill_if_none(s.start, 0)
        stop = fill_if_none(s.stop, n)
        start, stop = [x if x >= 0 else n+x for x in (start, stop)]
        return slice(start, stop, s.step)

    if not isinstance(shape, collections.abc.Sequence):
        shape = (shape,)

    if isinstance(roi, slice):
        return norm_slice(roi, shape[0])

    return tuple([norm_slice(s, n) for s, n in zip(roi, shape)])


def roi_pad(roi, pad, shape):
    """
    Pad ROI on each side, with clamping (0,..) -> shape
    """
    def pad_slice(s, n):
        return slice(max(0, s.start - pad), min(n, s.stop + pad))

    if isinstance(roi, slice):
        return pad_slice(roi, shape)

    return tuple(pad_slice(s, n) for s, n in zip(roi, shape))


def apply_affine(A: Affine, x: np.ndarray, y: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
    """
    broadcast A*(x_i, y_i) across all elements of x/y arrays in any shape (usually 2d image)
    """

    shape = x.shape

    A = np.asarray(A).reshape(3, 3)  # type: ignore[assignment]
    t = A[:2, -1].reshape((2, 1))    # type: ignore[index]
    A = A[:2, :2]                    # type: ignore[index]

    x, y = A @ np.vstack([x.ravel(), y.ravel()]) + t
    x, y = (a.reshape(shape) for a in (x, y))
    return (x, y)


def split_translation(t):
    """
    Split translation into pixel aligned and sub-pixel components.

    Subpixel translation is guaranteed to be in [-0.5, +0.5] range.

    >  x + t = x + t_whole + t_subpix

    :param t: (float, float)

    :returns: (t_whole: (float, float), t_subpix: (float, float))
    """
    from math import fmod

    def _split1(x):
        x_part = fmod(x, 1.0)
        x_whole = x - x_part
        if x_part > 0.5:
            x_part -= 1
            x_whole += 1
        elif x_part < -0.5:
            x_part += 1
            x_whole -= 1

        return (x_whole, x_part)

    _tt = [_split1(x) for x in t]

    return tuple(t[0] for t in _tt), tuple(t[1] for t in _tt)


def is_affine_st(A, tol=1e-10):
    """
    True if Affine transform has scale and translation components only.
    """
    (_, wx, _,
     wy, _, _,
     *_) = A

    return abs(wx) < tol and abs(wy) < tol


def decompose_rws(A):
    """
    Compute decomposition Affine matrix sans translation into Rotation, Shear and Scale.

    Note: that there are ambiguities for negative scales.

    Example: R(90)*S(1,1) == R(-90)*S(-1,-1),
    (R*(-I))*((-I)*S) == R*S

    A = R W S

    Where:

    R [ca -sa]  W [1, w]  S [sx,  0]
      [sa  ca]    [0, 1]    [ 0, sy]

    :return: Rotation, Sheer, Scale
    """
    # pylint: disable=too-many-locals

    from numpy.linalg import cholesky, det, inv

    if isinstance(A, Affine):
        def to_affine(m, t=(0, 0)):
            a, b, d, e = m.ravel()
            c, f = t
            return Affine(a, b, c,
                          d, e, f)

        (a, b, c,
         d, e, f,
         *_) = A
        R, W, S = decompose_rws(np.asarray([[a, b],
                                            [d, e]], dtype='float64'))

        return to_affine(R, (c, f)), to_affine(W), to_affine(S)

    assert A.shape == (2, 2)

    WS = cholesky(A.T @ A).T
    R = A @ inv(WS)

    if det(R) < 0:
        R[:, -1] *= -1
        WS[-1, :] *= -1

    ss = np.diag(WS)
    S = np.diag(ss)
    W = WS @ np.diag(1.0/ss)

    return R, W, S


def affine_from_pts(X, Y):
    """
    Given points X,Y compute A, such that: Y = A*X.

    Needs at least 3 points.

    :rtype: Affine
    """
    from numpy.linalg import lstsq

    assert len(X) == len(Y)
    assert len(X) >= 3

    n = len(X)

    XX = np.ones((n, 3), dtype='float64')
    YY = np.vstack(Y)
    for i, x in enumerate(X):
        XX[i, :2] = x

    mm, *_ = lstsq(XX, YY, rcond=-1)
    a, d, b, e, c, f = mm.ravel()

    return Affine(a, b, c,
                  d, e, f)


def get_scale_from_linear_transform(A):
    """
    Given a linear transform compute scale change.

    1. Y = A*X + t
    2. Extract scale components of A

    Returns (sx, sy), where sx > 0, sy > 0
    """
    _, _, S = decompose_rws(A)
    return abs(S.a), abs(S.e)


def get_scale_at_point(pt, tr, r=None):
    """
    Given an arbitrary locally linear transform estimate scale change around a point.

    1. Approximate Y = tr(X) as Y = A*X+t in the neighbourhood of pt, for X,Y in R2
    2. Extract scale components of A


    pt - estimate transform around this point
    r  - radius around the point (default 1)

    tr - List((x,y)) -> List((x,y))
         takes list of 2-d points on input and outputs same length list of 2d on output

    Returns (sx, sy), where sx > 0, sy > 0
    """
    pts0 = [(0, 0), (-1, 0), (0, -1), (1, 0), (0, 1)]
    x0, y0 = pt
    if r is None:
        XX = [(float(x+x0), float(y+y0)) for x, y in pts0]
    else:
        XX = [(float(x*r+x0), float(y*r+y0)) for x, y in pts0]
    YY = tr(XX)
    A = affine_from_pts(XX, YY)
    return get_scale_from_linear_transform(A)


def _same_crs_pix_transform(src, dst):
    assert src.crs == dst.crs

    def transform(pts, A):
        return [A*pt[:2] for pt in pts]

    _fwd = (~dst.transform) * src.transform  # src -> dst
    _bwd = ~_fwd                             # dst -> src

    def pt_tr(pts):
        return transform(pts, _fwd)
    pt_tr.back = lambda pts: transform(pts, _bwd)
    pt_tr.back.back = pt_tr
    pt_tr.linear = _fwd
    pt_tr.back.linear = _bwd

    return pt_tr


def compute_axis_overlap(Ns: int, Nd: int, s: float, t: float) -> Tuple[slice, slice]:  # noqa: N803
    """
    s, t define linear transform from destination coordinate space to source
    >>  x_s = s * x_d + t

    Ns -- number of pixels along some dimension of source image: (0, Ns)
    Nd -- same as Ns but for destination image

    :returns: (slice in the source image,
               slice in the destination image)
    """
    from math import floor, ceil

    needs_flip = s < 0

    if needs_flip:
        # change s, t to map into flipped src, i.e. src[::-1]
        s, t = -s, Ns - t

    assert s > 0

    # x_d = (x_s - t)/s => 1/s * x_s + t*(-1/s)
    #
    # x_d = s_ * x_s + t_
    s_ = 1.0/s
    t_ = -t*s_

    if t < 0:
        #  |<------- ... D
        #      |<--- ... S
        _in = (0, min(floor(t_), Nd))
    else:
        #        |<--... D
        # |<---------... S
        _in = (min(floor(t), Ns), 0)

    a = ceil(Nd*s + t)
    if a <= Ns:
        # ...----->|    D
        # ...-------->| S
        _out = (max(a, 0), Nd)
    else:
        # ...-------->|  D
        # ...----->|     S
        _out = (Ns, max(0, ceil(Ns*s_ + t_)))

    src, dst = (slice(_in[i], _out[i]) for i in range(2))

    if needs_flip:
        # remap src from flipped space to normal
        src = slice(Ns - src.stop, Ns - src.start)  # type: ignore

    return (src, dst)


def box_overlap(src_shape, dst_shape, ST, tol):  # noqa: N803
    """
    Given two image planes whose coordinate systems are related via scale and
    translation only, find overlapping regions within both.

    :param src_shape: Shape of source image plane
    :param dst_shape: Shape of destination image plane
    :param        ST: Affine transform with only scale/translation,
                      direction is: Xsrc = ST*Xdst
    :param       tol: Sub-pixel translation tolerance that's scaled by resolution.
    """
    from datacube.utils.math import maybe_int, snap_scale

    (sx, _, tx,
     _, sy, ty,
     *_) = ST

    sy = snap_scale(sy)
    sx = snap_scale(sx)

    ty = maybe_int(ty, tol)
    tx = maybe_int(tx, tol)

    s0, d0 = compute_axis_overlap(src_shape[0], dst_shape[0], sy, ty)
    s1, d1 = compute_axis_overlap(src_shape[1], dst_shape[1], sx, tx)
    return (s0, s1), (d0, d1)


def native_pix_transform(src, dst):
    """

    direction: from src to dst
    .back: goes the other way
    .linear: None|Affine linear transform src->dst if transform is linear (i.e. same CRS)
    """
    # Special case CRS_in == CRS_out
    if src.crs == dst.crs:
        return _same_crs_pix_transform(src, dst)

    _in = SimpleNamespace(crs=src.crs, A=src.transform)
    _out = SimpleNamespace(crs=dst.crs, A=dst.transform)

    _fwd = _in.crs.transformer_to_crs(_out.crs)
    _bwd = _out.crs.transformer_to_crs(_in.crs)

    _fwd = (_in.A, _fwd, ~_out.A)
    _bwd = (_out.A, _bwd, ~_in.A)

    def transform(pts, params):
        A, f, B = params
        return [B*pt[:2] for pt in [f(*(A*pt[:2])) for pt in pts]]

    def tr(pts):
        return transform(pts, _fwd)
    tr.back = lambda pts: transform(pts, _bwd)
    tr.back.back = tr
    tr.linear = None
    tr.back.linear = None

    return tr


def roi_intersect(a, b):
    """
    Compute intersection of two ROIs
    """
    def slice_intersect(a, b):
        if a.stop < b.start:
            return slice(a.stop, a.stop)
        elif a.start > b.stop:
            return slice(a.start, a.start)

        _in = max(a.start, b.start)
        _out = min(a.stop, b.stop)

        return slice(_in, _out)

    if isinstance(a, slice):
        if not isinstance(b, slice):
            b = b[0]
        return slice_intersect(a, b)

    b = (b,) if isinstance(b, slice) else b

    return tuple(slice_intersect(sa, sb) for sa, sb in zip(a, b))


def roi_center(roi):
    """ Return center point of roi
    """
    def slice_center(s):
        return (s.start + s.stop)*0.5

    if isinstance(roi, slice):
        return slice_center(roi)

    return tuple(slice_center(s) for s in roi)


def roi_from_points(xy, shape, padding=0, align=None):
    """
    Compute envelope around a bunch of points and return it as roi (tuple of
    row/col slices)

    Returned roi is clipped (0,0) --> shape, so it won't stick outside of the
    valid region.
    """
    def to_roi(*args):
        return tuple(slice(v[0], v[1]) for v in args)

    assert len(shape) == 2
    assert xy.ndim == 2 and xy.shape[1] == 2

    ny, nx = shape

    _in = np.floor(xy.min(axis=0)).astype('int32') - padding
    _out = np.ceil(xy.max(axis=0)).astype('int32') + padding

    if align is not None:
        _in = align_down(_in, align)
        _out = align_up(_out, align)

    xx = np.asarray([_in[0], _out[0]])
    yy = np.asarray([_in[1], _out[1]])

    xx = np.clip(xx, 0, nx, out=xx)
    yy = np.clip(yy, 0, ny, out=yy)

    return to_roi(yy, xx)


def compute_reproject_roi(src, dst, tol=0.05, padding=None, align=None):
    """Given two GeoBoxes find the region within the source GeoBox that overlaps
    with the destination GeoBox, and also compute the scale factor (>1 means
    shrink). Scale is chosen such that if you apply it to the source image
    before reprojecting, then reproject will have roughly no scale component.

    So we breaking up reprojection into two stages:

    1. Scale in the native pixel CRS
    2. Reprojection (possibly non-linear with CRS change)

    - src[roi] -> scale      -> reproject -> dst  (using native pixels)
    - src(scale)[roi(scale)] -> reproject -> dst  (using overview image)

    Here roi is "minimal", padding is configurable though, so you only read what you need.
    Also scale can be used to pick the right kind of overview level to read.

    Applying reprojection in two steps allows us to use pre-computed overviews,
    particularly useful when shrink factor is large. But even for data sources
    without overviews there are advantages for shrinking source image before
    applying reprojection: mainly quality of the output (reduces aliasing for
    large shrink factors), improved efficiency of the computation is likely as
    well.

    Also compute and return ROI of the dst geobox that is affected by src.

    If padding is None "appropriate" padding will be used depending on the
    transform between src<>dst:

    - No padding beyond sub-pixel alignment if Scale+Translation
    - 1 pixel source padding in all other cases

    :param tol: Sub-pixel translation tolerance as a percentage of resolution.

    :returns: SimpleNamespace with following fields:
     .roi_src    : (slice, slice)
     .roi_dst    : (slice, slice)
     .scale      : float
     .scale2     : (sx: float, sy: float)
     .is_st      : True|False
     .transform  : src coord -> dst coord

    For scale direction is: "scale > 1 --> shrink src to fit dst"

    """
    pts_per_side = 5

    def compute_roi(src, dst, tr, pts_per_side, padding, align):
        XY = np.vstack(tr.back(gbox_boundary(dst, pts_per_side)))
        roi_src = roi_from_points(XY, src.shape, padding, align=align)

        if roi_is_empty(roi_src):
            return (roi_src, np.s_[0:0, 0:0])

        # project src roi back into dst and compute roi from that
        xy = np.vstack(tr(roi_boundary(roi_src, pts_per_side)))
        roi_dst = roi_from_points(xy, dst.shape, padding=0)  # no need to add padding twice
        return (roi_src, roi_dst)

    tr = native_pix_transform(src, dst)

    if tr.linear is not None:
        tight_ok = align in (None, 0) and padding in (0, None)
        is_st = is_affine_st(tr.linear)

        if tight_ok and is_st:
            roi_src, roi_dst = box_overlap(src.shape, dst.shape, tr.back.linear, tol)
        else:
            padding = 1 if padding is None else padding
            roi_src, roi_dst = compute_roi(src, dst, tr, 2, padding, align)

        scale2 = get_scale_from_linear_transform(tr.linear)
    else:
        is_st = False
        padding = 1 if padding is None else padding

        roi_src, roi_dst = compute_roi(src, dst, tr, pts_per_side, padding, align)
        center_pt = roi_center(roi_src)[::-1]
        scale2 = get_scale_at_point(center_pt, tr)

    # change scale direction to be a shrink by factor
    scale2 = tuple(1/s for s in scale2)
    scale = min(scale2)

    return SimpleNamespace(roi_src=roi_src,
                           roi_dst=roi_dst,
                           scale=scale,
                           scale2=scale2,
                           is_st=is_st,
                           transform=tr)



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2020 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
import datetime
import pandas
import numpy as np
from types import SimpleNamespace

import pytest

from datacube.api.query import Query, _datetime_to_timestamp, query_group_by, solar_day, GroupBy, solar_offset
from datacube.model import Range
from datacube.utils import parse_time
from datacube.utils.geometry import CRS


@pytest.fixture
def mock_index():
    from unittest.mock import MagicMock
    return MagicMock()


def test_datetime_to_timestamp():
    assert _datetime_to_timestamp((1990, 1, 7)) == 631670400
    assert _datetime_to_timestamp(datetime.datetime(1990, 1, 7)) == 631670400
    assert _datetime_to_timestamp(631670400) == 631670400
    assert _datetime_to_timestamp('1990-01-07T00:00:00.0Z') == 631670400


def test_query_kwargs(mock_index):
    mock_index.datasets.get_field_names = lambda: {u'product', u'lat', u'sat_path', 'type_id', u'time', u'lon',
                                                   u'orbit', u'instrument', u'sat_row', u'platform', 'metadata_type',
                                                   u'gsi', 'type', 'id'}

    query = Query(index=mock_index, product='ls5_nbar_albers')
    assert str(query)
    assert query.product == 'ls5_nbar_albers'
    assert query.search_terms['product'] == 'ls5_nbar_albers'

    query = Query(index=mock_index, latitude=(-35, -36), longitude=(148, 149))
    assert query.geopolygon
    assert 'lat' in query.search_terms
    assert 'lon' in query.search_terms

    query = Query(index=mock_index, latitude=-35, longitude=148)
    assert query.geopolygon
    assert 'lat' in query.search_terms
    assert 'lon' in query.search_terms

    query = Query(index=mock_index, y=(-4174726, -4180011), x=(1515184, 1523263), crs='EPSG:3577')
    assert query.geopolygon
    assert 'lat' in query.search_terms
    assert 'lon' in query.search_terms

    query = Query(index=mock_index, y=-4174726, x=1515184, crs='EPSG:3577')
    assert query.geopolygon
    assert 'lat' in query.search_terms
    assert 'lon' in query.search_terms

    query = Query(index=mock_index, y=-4174726, x=1515184, crs='EPSG:3577')
    assert query.geopolygon
    assert 'lat' in query.search_terms
    assert 'lon' in query.search_terms

    query = Query(index=mock_index, y=-4174726, x=1515184, crs=CRS('EPSG:3577'))
    assert query.geopolygon
    assert 'lat' in query.search_terms
    assert 'lon' in query.search_terms

    query = Query(index=mock_index, time='2001')
    assert 'time' in query.search

    query = Query(index=mock_index, time=('2001', '2002'))
    assert 'time' in query.search

    with pytest.raises(ValueError):
        query = Query(index=mock_index, time=('2001', '2002', '2003'))

    with pytest.raises(ValueError):
        query = Query(index=mock_index, time=['2001', '2002', '2003'])

    with pytest.raises(ValueError):
        Query(index=mock_index,
              y=-4174726, coordinate_reference_system='WGS84',
              x=1515184, crs='EPSG:3577')

    with pytest.raises(LookupError):
        Query(index=mock_index, y=-4174726, x=1515184, crs='EPSG:3577', made_up_key='NotReal')

    with pytest.raises(LookupError):
        query_group_by(group_by='magic')

    gb = query_group_by('time')
    assert isinstance(gb, GroupBy)
    assert query_group_by(group_by=gb) is gb


def format_test(start_out, end_out):
    return Range(pandas.to_datetime(start_out, utc=True).to_pydatetime(),
                 pandas.to_datetime(end_out, utc=True).to_pydatetime())


testdata = [
    ((datetime.datetime(2008, 1, 1), datetime.datetime(2008, 1, 10)),
     format_test('2008-01-01T00:00:00', '2008-01-10T00:00:00.999999')),
    ((datetime.datetime(2008, 1, 1), datetime.datetime(2008, 1, 10, 23, 0, 0)),
     format_test('2008-01-01T00:00:00', '2008-01-10T23:00:00.999999')),
    ((datetime.datetime(2008, 1, 1), datetime.datetime(2008, 1, 10, 23, 59, 40)),
     format_test('2008-01-01T00:00:00', '2008-01-10T23:59:40.999999')),
    (('2008'),
     format_test('2008-01-01T00:00:00', '2008-12-31T23:59:59.999999')),
    (('2008', '2008'),
     format_test('2008-01-01T00:00:00', '2008-12-31T23:59:59.999999')),
    (('2008', '2009'),
     format_test('2008-01-01T00:00:00', '2009-12-31T23:59:59.999999')),
    (('2008-03', '2009'),
     format_test('2008-03-01T00:00', '2009-12-31T23:59:59.999999')),
    (('2008-03', '2009-10'),
     format_test('2008-03-01T00:00', '2009-10-31T23:59:59.999999')),
    (('2008', '2009-10'),
     format_test('2008-01-01T00:00', '2009-10-31T23:59:59.999999')),
    (('2008-03-03', '2008-11'),
     format_test('2008-03-03T00:00:00', '2008-11-30T23:59:59.999999')),
    (('2008-11-14', '2008-11-30'),
     format_test('2008-11-14T00:00:00', '2008-11-30T23:59:59.999999')),
    (('2008-11-14', '2008-11-29'),
     format_test('2008-11-14T00:00:00', '2008-11-29T23:59:59.999999')),
    (('2008-11-14', '2008-11'),
     format_test('2008-11-14T00:00:00', '2008-11-30T23:59:59.999999')),
    (('2008-11-14', '2008'),
     format_test('2008-11-14T00:00:00', '2008-12-31T23:59:59.999999')),
    (('2008-11-14'),
     format_test('2008-11-14T00:00:00', '2008-11-14T23:59:59.999999')),
    (('2008-11-14', '2009-02-02'),
     format_test('2008-11-14T00:00:00', '2009-02-02T23:59:59.999999')),
    (('2008-11-14T23:33:57', '2008-11-14 23:33:57'),
     format_test('2008-11-14T23:33:57', '2008-11-14T23:33:57.999999')),
    (('2008-11-14 23:33', '2008-11-14 23:34'),
     format_test('2008-11-14T23:33:00', '2008-11-14T23:34:59.999999')),
    (('2008-11-14T23:00:00', '2008-11-14 23:35'),
     format_test('2008-11-14T23:00', '2008-11-14T23:35:59.999999')),
    (('2008-11-10T11', '2008-11-16 14:01'),
     format_test('2008-11-10T11:00', '2008-11-16T14:01:59.999999')),
    ((datetime.date(1995, 1, 1), datetime.date(1999, 1, 1)),
     format_test('1995-01-01T00:00:00', '1999-01-01T23:59:59.999999')),
    ((datetime.datetime(2008, 1, 1), datetime.datetime(2008, 1, 10, 23, 59, 40)),
     format_test('2008-01-01T00:00:00', '2008-01-10T23:59:40.999999')),
    ((datetime.date(2008, 1, 1)),
     format_test('2008-01-01T00:00:00', '2008-01-01T23:59:59.999999'))
]


@pytest.mark.parametrize('time_param,expected', testdata)
def test_time_handling(time_param, expected):
    query = Query(time=time_param)
    assert 'time' in query.search_terms
    assert query.search_terms['time'] == expected


def test_solar_day():
    _s = SimpleNamespace
    ds = _s(center_time=parse_time('1987-05-22 23:07:44.2270250Z'),
            metadata=_s(lon=Range(begin=150.415,
                                  end=152.975)))

    assert solar_day(ds) == np.datetime64('1987-05-23', 'D')
    assert solar_day(ds, longitude=0) == np.datetime64('1987-05-22', 'D')

    ds.metadata = _s()

    with pytest.raises(ValueError) as e:
        solar_day(ds)

    assert 'Cannot compute solar_day: dataset is missing spatial info' in str(e.value)


def test_solar_offset():
    from datacube.utils.geometry import point
    from datetime import timedelta

    def _hr(t):
        return t.days*24 + t.seconds/3600

    def p(lon):
        return point(lon, 0, 'epsg:4326')

    assert solar_offset(p(0)) == timedelta(seconds=0)
    assert solar_offset(p(0).to_crs('epsg:3857')) == timedelta(seconds=0)

    assert solar_offset(p(179.9)) == timedelta(hours=12)
    assert _hr(solar_offset(p(-179.9))) == -12.0

    assert solar_offset(p(20), 's') != solar_offset(p(20), 'h')
    assert solar_offset(p(20), 's') < solar_offset(p(21), 's')

    _s = SimpleNamespace
    ds = _s(center_time=parse_time('1987-05-22 23:07:44.2270250Z'),
            metadata=_s(lon=Range(begin=150.415,
                                  end=152.975)))
    assert solar_offset(ds) == timedelta(hours=10)
    ds.metadata = _s()

    with pytest.raises(ValueError):
        solar_offset(ds)


def test_dateline_query_building():
    lon = Query(x=(618300, 849000),
                y=(-1876800, -1642500),
                crs='EPSG:32660').search_terms['lon']

    assert lon.begin < 180 < lon.end


def test_query_issue_1146():
    q = Query(k='AB')
    assert q.search['k'] == 'AB'


def test_query_multiple_products(mock_index):
    q = Query(index=mock_index, product=['ls5_nbar_albers', 'ls7_nbar_albers'])
    assert q.product == ['ls5_nbar_albers', 'ls7_nbar_albers']



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2020 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
import datetime
import os
import platform
import sys
import uuid
import toolz

import numpy
import xarray
import yaml
from pandas import to_datetime

import datacube
from datacube.model import Dataset
from datacube.utils import geometry, SimpleDocNav, InvalidDocException
from datacube.utils.py import sorted_items

try:
    from yaml import CSafeDumper as SafeDumper  # type: ignore
except ImportError:
    from yaml import SafeDumper  # type: ignore


class BadMatch(Exception):  # noqa: N818
    pass


def machine_info():
    info = {
        'software_versions': {
            'python': {'version': sys.version},
            'datacube': {'version': datacube.__version__,
                         'repo_url': 'https://github.com/opendatacube/datacube-core.git'},
        },
        'hostname': platform.node(),
    }

    if hasattr(os, 'uname'):
        info['uname'] = ' '.join(os.uname())
    else:
        info['uname'] = ' '.join([platform.system(),
                                  platform.node(),
                                  platform.release(),
                                  platform.version(),
                                  platform.machine()])

    return {'lineage': {'machine': info}}


def geobox_info(extent, valid_data=None):
    image_bounds = extent.boundingbox
    data_bounds = valid_data.boundingbox if valid_data else image_bounds
    ul = geometry.point(data_bounds.left, data_bounds.top, crs=extent.crs).to_crs(geometry.CRS('EPSG:4326'))
    ur = geometry.point(data_bounds.right, data_bounds.top, crs=extent.crs).to_crs(geometry.CRS('EPSG:4326'))
    lr = geometry.point(data_bounds.right, data_bounds.bottom, crs=extent.crs).to_crs(geometry.CRS('EPSG:4326'))
    ll = geometry.point(data_bounds.left, data_bounds.bottom, crs=extent.crs).to_crs(geometry.CRS('EPSG:4326'))
    doc = {
        'extent': {
            'coord': {
                'ul': {'lon': ul.points[0][0], 'lat': ul.points[0][1]},
                'ur': {'lon': ur.points[0][0], 'lat': ur.points[0][1]},
                'lr': {'lon': lr.points[0][0], 'lat': lr.points[0][1]},
                'll': {'lon': ll.points[0][0], 'lat': ll.points[0][1]},
            }
        },
        'grid_spatial': {
            'projection': {
                'spatial_reference': str(extent.crs),
                'geo_ref_points': {
                    'ul': {'x': image_bounds.left, 'y': image_bounds.top},
                    'ur': {'x': image_bounds.right, 'y': image_bounds.top},
                    'll': {'x': image_bounds.left, 'y': image_bounds.bottom},
                    'lr': {'x': image_bounds.right, 'y': image_bounds.bottom},
                }
            }
        }
    }
    if valid_data:
        doc['grid_spatial']['projection']['valid_data'] = valid_data.__geo_interface__
    return doc


def new_dataset_info():
    return {
        'id': str(uuid.uuid4()),
        'creation_dt': datetime.datetime.utcnow().isoformat(),
    }


def band_info(band_names, band_uris=None):
    """
    :param list band_names: names of the bands
    :param dict band_uris: mapping from names to dicts with 'path' and 'layer' specs
    """
    if band_uris is None:
        band_uris = {name: {'path': '', 'layer': name} for name in band_names}

    return {
        'image': {
            'bands': {name: band_uris[name] for name in band_names}
        }
    }


def time_info(time, start_time=None, end_time=None, key_time=None):
    time_str = to_datetime(time).isoformat()
    start_time_str = to_datetime(start_time).isoformat() if start_time else time_str
    end_time_str = to_datetime(end_time).isoformat() if end_time else time_str
    extent = {
        'extent': {
            'from_dt': start_time_str,
            'to_dt': end_time_str,
            'center_dt': time_str,
        }
    }
    if key_time is not None:
        extent['extent']['key_time'] = to_datetime(key_time).isoformat()
    return extent


def source_info(source_datasets):
    return {
        'lineage': {
            'source_datasets': {str(idx): dataset.metadata_doc for idx, dataset in enumerate(source_datasets)}
        }
    }


def datasets_to_doc(output_datasets):
    """
    Create a yaml document version of every dataset

    :param output_datasets: An array of :class:`datacube.model.Dataset`
    :type output_datasets: :py:class:`xarray.DataArray`
    :return: An array of yaml document strings
    :rtype: :py:class:`xarray.DataArray`
    """

    def dataset_to_yaml(index, dataset):
        return yaml.dump(dataset.metadata_doc, Dumper=SafeDumper, encoding='utf-8')

    return xr_apply(output_datasets, dataset_to_yaml, dtype='O').astype('S')


def xr_iter(data_array):
    """
    Iterate over every element in an xarray, returning::

        * the numerical index eg ``(10, 1)``
        * the labeled index eg ``{'time': datetime(), 'band': 'red'}``
        * the element (same as ``da[10, 1].item()``)

    :param data_array: Array to iterate over
    :type data_array: xarray.DataArray
    :return: i-index, label-index, value of da element
    :rtype tuple, dict, da.dtype
    """
    values = data_array.values
    coords = {coord_name: v.values for coord_name, v in data_array.coords.items()}
    for i in numpy.ndindex(data_array.shape):
        entry = values[i]
        index = {coord_name: v[i] for coord_name, v in coords.items()}
        yield i, index, entry


def xr_apply(data_array, func, dtype=None, with_numeric_index=False):
    """
    Apply a function to every element of a :class:`xarray.DataArray`

    :type data_array: xarray.DataArray
    :param func: function that takes a dict of labels and an element of the array,
        and returns a value of the given dtype
    :param dtype: The dtype of the returned array, default to the same as original
    :param with_numeric_index Bool: If true include numeric index: func(index, labels, value)
    :return: The array with output of the function for every element.
    :rtype: xarray.DataArray
    """
    if dtype is None:
        dtype = data_array.dtype

    data = numpy.empty(shape=data_array.shape, dtype=dtype)
    for i, index, entry in xr_iter(data_array):
        if with_numeric_index:
            v = func(i, index, entry)
        else:
            v = func(index, entry)
        data[i] = v
    return xarray.DataArray(data, coords=data_array.coords, dims=data_array.dims)


def make_dataset(product, sources, extent, center_time, valid_data=None, uri=None, app_info=None,
                 band_uris=None, start_time=None, end_time=None):
    """
    Create :class:`datacube.model.Dataset` for the data

    :param DatasetType product: Product the dataset is part of
    :param list[:class:`Dataset`] sources: datasets used to produce the dataset
    :param Geometry extent: extent of the dataset
    :param Geometry valid_data: extent of the valid data
    :param center_time: time of the central point of the dataset
    :param str uri: The uri of the dataset
    :param dict app_info: Additional metadata to be stored about the generation of the product
    :param dict band_uris: band name to uri mapping
    :param start_time: start time of the dataset (defaults to `center_time`)
    :param end_time: end time of the dataset (defaults to `center_time`)
    :rtype: class:`Dataset`
    """
    document = {}
    merge(document, product.metadata_doc)
    merge(document, new_dataset_info())
    merge(document, machine_info())
    merge(document, band_info(product.measurements.keys(), band_uris=band_uris))
    merge(document, source_info(sources))
    merge(document, geobox_info(extent, valid_data))
    merge(document, time_info(center_time, start_time, end_time))
    merge(document, app_info or {})

    return Dataset(product,
                   document,
                   uris=[uri] if uri else None,
                   sources={str(idx): dataset for idx, dataset in enumerate(sources)})


def merge(a, b, path=None):
    """
    Merge dictionary `b` into dictionary `a`

    See: http://stackoverflow.com/a/7205107/5262498

    :type a: dict
    :type b: dict
    :rtype: dict
    """
    if path is None:
        path = []
    for key in b:
        if key in a:
            if isinstance(a[key], dict) and isinstance(b[key], dict):
                merge(a[key], b[key], path + [str(key)])
            elif a[key] == b[key]:
                pass  # same leaf value
            else:
                raise Exception('Conflict at %s' % '.'.join(path + [str(key)]))
        else:
            a[key] = b[key]
    return a


def traverse_datasets(ds, cbk, mode='post-order', **kwargs):
    """Perform depth first traversal of lineage tree. Note that we assume it's a
    tree, even though it might be a DAG (Directed Acyclic Graph). If it is a
    DAG it will be treated as if it was a tree with some nodes appearing twice or more
    times in this tree.

    Order of traversal of nodes on the same level is in default sort order for
    strings (assuming your keys are strings, which is the case for Dataset
    object). NOTE: this could be problematic as order might be dependent on
    locale settings.

    If given a graph with cycles this will blow the stack, so don't do that.

    ds -- Dataset with lineage to iterate over, but really anything that has
          `sources` attribute which contains a dict from string to the same
          thing.

    cbk :: (Dataset, depth=0, name=None, **kwargs) -> None

    mode: post-order | pre-order

    mode=post-order -- Visit all lineage first, only then visit top level
    mode=pre-order --  Visit top level first, only then visit lineage

    """

    def visit_pre_order(ds, func, depth=0, name=None):
        func(ds, depth=depth, name=name, **kwargs)

        for k, v in sorted_items(ds.sources):
            visit_pre_order(v, func, depth=depth+1, name=k)

    def visit_post_order(ds, func, depth=0, name=None):
        for k, v in sorted_items(ds.sources):
            visit_post_order(v, func, depth=depth+1, name=k)

        func(ds, depth=depth, name=name, **kwargs)

    proc = {'post-order': visit_post_order,
            'pre-order': visit_pre_order}.get(mode, None)

    if proc is None:
        raise ValueError('Unsupported traversal mode: {}'.format(mode))

    proc(ds, cbk)


def flatten_datasets(ds, with_depth_grouping=False):
    """Build a dictionary mapping from dataset.id to a list of datasets with that
    id appearing in the lineage DAG. When DAG is unrolled into a tree, some
    datasets will be reachable by multiple paths, sometimes these would be
    exactly the same python object, other times they will be duplicate views of
    the same "conceptual dataset object". If the same dataset is reachable by
    three possible paths from the root, it will appear three times in the
    flattened view.

    ds could be a Dataset object read from DB with `include_sources=True`, or
    it could be `SimpleDocNav` object created from a dataset metadata document
    read from a file.

    If with_depth_grouping=True, also build depth -> [Ds] mapping and return it
    along with Id -> [Ds] mapping. In this case top level is depth=0.
    """
    def get_list(out, k):
        if k not in out:
            out[k] = []
        return out[k]

    def proc(ds, depth=0, name=None, id_map=None, depth_map=None):
        k = ds.id

        get_list(id_map, k).append(ds)
        if depth_map is not None:
            get_list(depth_map, depth).append(ds)

    id_map = {}
    depth_map = {} if with_depth_grouping else None

    traverse_datasets(ds, proc, id_map=id_map, depth_map=depth_map)

    if depth_map:
        # convert dict Int->V to just a list
        dout = [None]*len(depth_map)
        for k, v in depth_map.items():
            dout[k] = v

        return id_map, dout

    return id_map


def remap_lineage_doc(root, mk_node, **kwargs):
    def visit(ds):
        return mk_node(ds,
                       {k: visit(v) for k, v in sorted_items(ds.sources)},
                       **kwargs)

    if not isinstance(root, SimpleDocNav):
        root = SimpleDocNav(root)

    try:
        return visit(root)
    except BadMatch as e:
        if str(root.id) not in str(e):
            raise BadMatch(f"Error loading lineage dataset: {e}") from None
        else:
            raise


def dedup_lineage(root):
    """Find duplicate nodes in the lineage tree and replace them with references.

    Will raise `ValueError` when duplicate dataset (same uuid, but different
    path from root) has either conflicting metadata or conflicting lineage
    data.

    :param dict|SimpleDocNav root:

    Returns a new document that has the same structure as input document, but
    with duplicate entries now being aliases rather than copies.
    """

    def check_sources(a, b):
        """ True if two dictionaries contain same objects under the same names.
        same, not just equivalent.
        """
        if len(a) != len(b):
            return False

        for ((ak, av), (bk, bv)) in zip(sorted_items(a), sorted_items(b)):
            if ak != bk:
                return False
            if av is not bv:
                return False

        return True

    def mk_node(ds, sources, cache, sources_path):
        existing = cache.get(ds.id, None)
        doc = ds.doc_without_lineage_sources

        if existing is not None:
            _ds, _doc, _sources = existing

            if not check_sources(sources, _sources):
                raise InvalidDocException('Inconsistent lineage for repeated dataset with _id: {}'.format(ds.id))

            if doc != _doc:
                raise InvalidDocException('Inconsistent metadata for repeated dataset with _id: {}'.format(ds.id))

            return _ds

        out_ds = toolz.assoc_in(doc, sources_path, sources)
        cache[ds.id] = (out_ds, doc, sources)
        return out_ds

    if not isinstance(root, SimpleDocNav):
        root = SimpleDocNav(root)

    return remap_lineage_doc(root, mk_node, cache={}, sources_path=root.sources_path)



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2020 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
"""
Test creation of added/updated columns during
`datacube system init`
"""
import pytest

from datacube.drivers.postgres.sql import SCHEMA_NAME
from datacube.drivers.postgres import _schema


COLUMN_PRESENCE = """
select exists (select 1 from information_schema.columns
where table_schema='{schema}' and table_name='{table}' and column_name='{column}');
"""

DROP_COLUMN = """
alter table {schema}.{table} drop column {column}
"""

TRIGGER_PRESENCE = """
select tgname
from pg_trigger
where not tgisinternal
and tgrelid = '{schema}.{table}'::regclass;
"""


def check_column(conn, table_name: str, column_name: str) -> bool:
    column_result = conn.execute(
        COLUMN_PRESENCE.format(
            schema=SCHEMA_NAME, table=table_name, column=column_name
        )
    ).fetchone()
    return column_result == (True,)


def check_trigger(conn, table_name: str) -> bool:
    trigger_result = conn.execute(
        TRIGGER_PRESENCE.format(schema=SCHEMA_NAME, table=table_name)
    ).fetchone()
    if trigger_result is None:
        return False
    return 'row_update_time' in trigger_result[0]


def drop_column(conn, table: str, column: str):
    conn.execute(DROP_COLUMN.format(
        schema=SCHEMA_NAME, table=table, column=column))


@pytest.mark.parametrize('datacube_env_name', ('datacube', ))
def test_added_column(clirunner, uninitialised_postgres_db):
    # Run on an empty database.
    result = clirunner(["system", "init"])
    assert "Created." in result.output

    with uninitialised_postgres_db._connect() as connection:
        assert check_column(connection, _schema.METADATA_TYPE.name, "updated")
        assert not check_column(connection, _schema.METADATA_TYPE.name, "fake_column")
        assert check_column(connection, _schema.PRODUCT.name, "updated")
        assert not check_column(connection, _schema.PRODUCT.name, "fake_column")
        assert check_column(connection, _schema.DATASET.name, "updated")
        assert not check_column(connection, _schema.DATASET.name, "fake_column")
        assert check_column(connection, _schema.DATASET_LOCATION.name, "added")
        assert not check_column(connection, _schema.DATASET_LOCATION.name, "fake_column")

        # Check for triggers
        assert check_trigger(connection, _schema.METADATA_TYPE.name)
        assert check_trigger(connection, _schema.PRODUCT.name)
        assert check_trigger(connection, _schema.DATASET.name)
        assert not check_trigger(connection, _schema.DATASET_LOCATION.name)


@pytest.mark.parametrize('datacube_env_name', ('datacube', ))
def test_readd_column(clirunner, uninitialised_postgres_db):
    # Run on an empty database. drop columns and readd
    result = clirunner(["system", "init"])
    assert "Created." in result.output

    with uninitialised_postgres_db._connect() as connection:
        # Drop all the columns for an init rerun
        drop_column(connection, _schema.METADATA_TYPE.name, "updated")
        drop_column(connection, _schema.PRODUCT.name, "updated")
        drop_column(connection, _schema.DATASET.name, "updated")
        drop_column(connection, _schema.DATASET_LOCATION.name, "added")

        assert not check_column(connection, _schema.METADATA_TYPE.name, "updated")
        assert not check_column(connection, _schema.PRODUCT.name, "updated")
        assert not check_column(connection, _schema.DATASET.name, "updated")
        assert not check_column(connection, _schema.DATASET_LOCATION.name, "added")

    result = clirunner(["system", "init"])

    with uninitialised_postgres_db._connect() as connection:
        assert check_column(connection, _schema.METADATA_TYPE.name, "updated")
        assert check_column(connection, _schema.PRODUCT.name, "updated")
        assert check_column(connection, _schema.DATASET.name, "updated")
        assert check_column(connection, _schema.DATASET_LOCATION.name, "added")



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2020 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
import logging
import xarray
from collections import OrderedDict
import pandas as pd

from datacube.utils.geometry import intersects
from .query import Query, query_group_by
from .core import Datacube

_LOG = logging.getLogger(__name__)


def _fast_slice(array, indexers):
    data = array.values[indexers]
    dims = [dim for dim, indexer in zip(array.dims, indexers) if isinstance(indexer, slice)]
    variable = xarray.Variable(dims, data, attrs=array.attrs, fastpath=True)
    coords = OrderedDict((dim,
                          xarray.Variable((dim,),
                                          array.coords[dim].values[indexer],
                                          attrs=array.coords[dim].attrs,
                                          fastpath=True))
                         for dim, indexer in zip(array.dims, indexers) if isinstance(indexer, slice))
    return xarray.DataArray(variable, coords=coords, fastpath=True)


class Tile(object):
    """
    The Tile object holds a lightweight representation of a datacube result.

    It is produced by :meth:`.GridWorkflow.list_cells` or :meth:`.GridWorkflow.list_tiles`.

    The Tile object can be passed to :meth:`GridWorkflow.load` to be loaded into memory as
    an :class:`xarray.Dataset`.

    A portion of a tile can be created by using index notation. eg:

        tile[0:1, 0:1000, 0:1000]

    This can be used to load small portions of data into memory, instead of having to access
    the entire `Tile` at once.
    """

    def __init__(self, sources, geobox):
        """Create a Tile representing a dataset that can be loaded.

        :param xarray.DataArray sources: An array of non-spatial dimensions of the request, holding lists of
            datacube.storage.DatasetSource objects.
        :param model.GeoBox geobox: The spatial footprint of the Tile
        """
        self.sources = sources
        self.geobox = geobox

    @property
    def dims(self):
        """Names of the dimensions, eg ``('time', 'y', 'x')``
        :return: tuple(str)
        """
        return self.sources.dims + self.geobox.dimensions

    @property
    def shape(self):
        """Lengths of each dimension, eg ``(285, 4000, 4000)``
        :return: tuple(int)
        """
        return self.sources.shape + self.geobox.shape

    @property
    def product(self):
        """
        :rtype: datacube.model.DatasetType
        """
        return self.sources.values[0][0].type

    def __getitem__(self, chunk):
        sources = _fast_slice(self.sources, chunk[:len(self.sources.shape)])
        geobox = self.geobox[chunk[len(self.sources.shape):]]
        return Tile(sources, geobox)

    # TODO(csiro) Split on time range
    def split(self, dim, step=1):
        """
        Splits along a non-spatial dimension into Tile objects with a length of 1 or more in the `dim` dimension.

        :param dim: Name of the non-spatial dimension to split
        :param step: step size to split
        :return: tuple(key, Tile)
        """
        axis = self.dims.index(dim)
        indexer = [slice(None)] * len(self.dims)
        size = self.sources[dim].size
        for i in range(0, size, step):
            indexer[axis] = slice(i, min(size, i + step))
            yield self.sources[dim].values[i], self[tuple(indexer)]

    def split_by_time(self, freq='A', time_dim='time', **kwargs):
        """
        Splits along the `time` dimension, into periods, using pandas offsets, such as:
        :
            'A': Annual
            'Q': Quarter
            'M': Month
        See: http://pandas.pydata.org/pandas-docs/stable/timeseries.html?highlight=rollback#timeseries-offset-aliases

        :param freq: time series frequency
        :param time_dim: name of the time dimension
        :param kwargs: other keyword arguments passed to ``pandas.period_range``
        :return: Generator[tuple(str, Tile)] generator of the key string (eg '1994') and the slice of Tile
        """
        # extract first and last timestamps from the time axis, note this will
        # work with 1 element arrays as well
        start_range, end_range = self.sources[time_dim].data[[0, -1]]

        for p in pd.period_range(start=start_range,
                                 end=end_range,
                                 freq=freq,
                                 **kwargs):
            sources_slice = self.sources.loc[{time_dim: slice(p.start_time, p.end_time)}]
            yield str(p), Tile(sources=sources_slice, geobox=self.geobox)

    def __str__(self):
        return "Tile<sources={!r},\n\tgeobox={!r}>".format(self.sources, self.geobox)

    def __repr__(self):
        return self.__str__()


class GridWorkflow(object):
    """
    GridWorkflow deals with cell- and tile-based processing using a grid defining a projection and resolution.

    Use GridWorkflow to specify your desired output grid.  The methods :meth:`list_cells` and :meth:`list_tiles`
    query the index and return a dictionary of cell or tile keys, each mapping to a :class:`Tile` object.

    The :class:`.Tile` object can then be used to load the data without needing the index,
    and can be serialized for use with the `distributed` package.
    """

    def __init__(self, index, grid_spec=None, product=None):
        """
        Create a grid workflow tool.

        Either grid_spec or product must be supplied.

        :param datacube.index.Index index: The database index to use.
        :param GridSpec grid_spec: The grid projection and resolution
        :param str product: The name of an existing product, if no grid_spec is supplied.
        """
        self.index = index
        if grid_spec is None:
            if product is None:
                raise ValueError("Have to supply either grid_spec or product")

            product = self.index.products.get_by_name(product)
            if product is None:
                raise ValueError(f"No such product: {product}")

            if product.grid_spec is None:
                raise ValueError(f"Supplied product `{product}` does not have a gridspec")

            grid_spec = product.grid_spec
        self.grid_spec = grid_spec

        assert self.grid_spec is not None

    def cell_observations(self, cell_index=None, geopolygon=None, tile_buffer=None, **indexers):
        """
        List datasets, grouped by cell.

        :param datacube.utils.Geometry geopolygon:
            Only return observations with data inside polygon.
        :param (float,float) tile_buffer:
            buffer tiles by (y, x) in CRS units
        :param (int,int) cell_index:
            The cell index. E.g. (14, -40)
        :param indexers:
            Query to match the datasets, see :py:class:`datacube.api.query.Query`
        :return: Datsets grouped by cell index
        :rtype: dict[(int,int), list[:py:class:`datacube.model.Dataset`]]

        .. seealso::
            :meth:`datacube.Datacube.find_datasets`

            :class:`datacube.api.query.Query`
        """
        # pylint: disable=too-many-locals
        # TODO: split this method into 3: cell/polygon/unconstrained querying

        if tile_buffer is not None and geopolygon is not None:
            raise ValueError('Cannot process tile_buffering and geopolygon together.')
        cells = {}

        def add_dataset_to_cells(tile_index, tile_geobox, dataset_):
            cells.setdefault(tile_index, {'datasets': [], 'geobox': tile_geobox})['datasets'].append(dataset_)

        if cell_index:
            assert len(cell_index) == 2
            cell_index = tuple(cell_index)
            geobox = self.grid_spec.tile_geobox(cell_index)
            geobox = geobox.buffered(*tile_buffer) if tile_buffer else geobox

            datasets, query = self._find_datasets(geobox.extent, indexers)
            for dataset in datasets:
                if intersects(geobox.extent, dataset.extent.to_crs(self.grid_spec.crs)):
                    add_dataset_to_cells(cell_index, geobox, dataset)
            return cells
        else:
            datasets, query = self._find_datasets(geopolygon, indexers)
            geobox_cache = {}

            if query.geopolygon:
                # Get a rough region of tiles
                query_tiles = set(
                    tile_index for tile_index, tile_geobox in
                    self.grid_spec.tiles_from_geopolygon(query.geopolygon, geobox_cache=geobox_cache))

                for dataset in datasets:
                    # Go through our datasets and see which tiles each dataset produces, and whether they intersect
                    # our query geopolygon.
                    dataset_extent = dataset.extent.to_crs(self.grid_spec.crs)
                    bbox = dataset_extent.boundingbox
                    bbox = bbox.buffered(*tile_buffer) if tile_buffer else bbox

                    for tile_index, tile_geobox in self.grid_spec.tiles(bbox, geobox_cache=geobox_cache):
                        if tile_index in query_tiles and intersects(tile_geobox.extent, dataset_extent):
                            add_dataset_to_cells(tile_index, tile_geobox, dataset)

            else:
                for dataset in datasets:
                    for tile_index, tile_geobox in self.grid_spec.tiles_from_geopolygon(dataset.extent,
                                                                                        tile_buffer=tile_buffer,
                                                                                        geobox_cache=geobox_cache):
                        add_dataset_to_cells(tile_index, tile_geobox, dataset)

            return cells

    def _find_datasets(self, geopolygon, indexers):
        query = Query(index=self.index, geopolygon=geopolygon, **indexers)
        if not query.product:
            raise RuntimeError('must specify a product')
        datasets = self.index.datasets.search_eager(**query.search_terms)
        return datasets, query

    @staticmethod
    def group_into_cells(observations, group_by):
        """
        Group observations into a stack of source tiles.

        :param observations: datasets grouped by cell index, like from :py:meth:`cell_observations`
        :param group_by: grouping method, as returned by :py:meth:`datacube.api.query.query_group_by`
        :type group_by: :py:class:`datacube.api.query.GroupBy`
        :return: tiles grouped by cell index
        :rtype: dict[(int,int), :class:`.Tile`]

        .. seealso::
            :meth:`load`

            :meth:`datacube.Datacube.group_datasets`
        """
        cells = {}
        for cell_index, observation in observations.items():
            sources = Datacube.group_datasets(observation['datasets'], group_by)
            cells[cell_index] = Tile(sources, observation['geobox'])
        return cells

    @staticmethod
    def tile_sources(observations, group_by):
        """
        Split observations into tiles and group into source tiles

        :param observations: datasets grouped by cell index, like from :meth:`cell_observations`
        :param group_by: grouping method, as returned by :py:meth:`datacube.api.query.query_group_by`
        :type group_by: :py:class:`datacube.api.query.GroupBy`
        :return: tiles grouped by cell index and time
        :rtype: dict[tuple(int, int, numpy.datetime64), :py:class:`.Tile`]

        .. seealso::
            :meth:`load`

            :meth:`datacube.Datacube.group_datasets`
        """
        tiles = {}
        for cell_index, observation in observations.items():
            dss = observation['datasets']
            geobox = observation['geobox']

            sources = Datacube.group_datasets(dss, group_by)
            coord = sources[sources.dims[0]]
            for i in range(coord.size):
                tile_index = cell_index + (coord.values[i],)
                tiles[tile_index] = Tile(sources[i:i+1], geobox)

        return tiles

    def list_cells(self, cell_index=None, **query):
        """
        List cells that match the query.

        Returns a dictionary of cell indexes to :class:`.Tile` objects.

        Cells are included if they contain any datasets that match the query using the same format as
        :meth:`datacube.Datacube.load`.

        E.g.::

            gw.list_cells(product='ls5_nbar_albers',
                          time=('2001-1-1 00:00:00', '2001-3-31 23:59:59'))

        :param (int,int) cell_index: The cell index. E.g. (14, -40)
        :param query: see :py:class:`datacube.api.query.Query`
        :rtype: dict[(int, int), :class:`.Tile`]
        """
        observations = self.cell_observations(cell_index, **query)
        return self.group_into_cells(observations, query_group_by(**query))

    def list_tiles(self, cell_index=None, **query):
        """
        List tiles of data, sorted by cell.
        ::

            tiles = gw.list_tiles(product='ls5_nbar_albers',
                                  time=('2001-1-1 00:00:00', '2001-3-31 23:59:59'))

        The values can be passed to :meth:`load`

        :param (int,int) cell_index: The cell index (optional). E.g. (14, -40)
        :param query: see :py:class:`datacube.api.query.Query`
        :rtype: dict[(int, int, numpy.datetime64), :class:`.Tile`]

        .. seealso:: :meth:`load`
        """
        observations = self.cell_observations(cell_index, **query)
        return self.tile_sources(observations, query_group_by(**query))

    @staticmethod
    def load(tile, measurements=None, dask_chunks=None, fuse_func=None, resampling=None, skip_broken_datasets=False):
        """
        Load data for a cell/tile.

        The data to be loaded is defined by the output of :meth:`list_tiles`.

        This is a static function and does not use the index. This can be useful when running as a worker in a
        distributed environment and you wish to minimize database connections.

        See the documentation on using `xarray with dask <http://xarray.pydata.org/en/stable/dask.html>`_
        for more information.

        :param `.Tile` tile: The tile to load.

        :param list(str) measurements: The names of measurements to load

        :param dict dask_chunks: If the data should be loaded as needed using :py:class:`dask.array.Array`,
            specify the chunk size in each output direction.

            See the documentation on using `xarray with dask <http://xarray.pydata.org/en/stable/dask.html>`_
            for more information.

        :param fuse_func: Function to fuse together a tile that has been pre-grouped by calling
            :meth:`list_cells` with a ``group_by`` parameter.

        :param str|dict resampling:

            The resampling method to use if re-projection is required, could be
            configured per band using a dictionary (:meth: `load_data`)

            Valid values are: ``'nearest', 'cubic', 'bilinear', 'cubic_spline', 'lanczos', 'average'``

            Defaults to ``'nearest'``.

        :param bool skip_broken_datasets: If True, ignore broken datasets and continue processing with the data
             that can be loaded. If False, an exception will be raised on a broken dataset. Defaults to False.

        :rtype: :py:class:`xarray.Dataset`

        .. seealso::
            :meth:`list_tiles` :meth:`list_cells`
        """
        measurement_dicts = tile.product.lookup_measurements(measurements)

        dataset = Datacube.load_data(tile.sources, tile.geobox,
                                     measurement_dicts, resampling=resampling,
                                     dask_chunks=dask_chunks, fuse_func=fuse_func,
                                     skip_broken_datasets=skip_broken_datasets)

        return dataset

    def update_tile_lineage(self, tile):
        for i in range(tile.sources.size):
            sources = tile.sources.values[i]
            tile.sources.values[i] = tuple(self.index.datasets.get(dataset.id, include_sources=True)
                                           for dataset in sources)
        return tile

    def __str__(self):
        return "GridWorkflow<index={!r},\n\tgridspec={!r}>".format(self.index, self.grid_spec)

    def __repr__(self):
        return self.__str__()



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2020 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
"""
Common methods for UI code.
"""
from pathlib import Path
from typing import Union, Optional

from toolz.functoolz import identity  # type: ignore[import]

from datacube.utils import read_documents, InvalidDocException, SimpleDocNav, is_supported_document_type, is_url


def get_metadata_path(possible_path: Union[str, Path]) -> str:
    """
    Find a metadata path for a given input/dataset path.

    Needs to handle local files as well as remote URLs
    """
    # We require exact URLs, lets skip any sort of fancy investigation and mapping
    if isinstance(possible_path, str) and is_url(possible_path):
        return possible_path

    dataset_path = Path(possible_path)

    # They may have given us a metadata file directly.
    if dataset_path.is_file() and is_supported_document_type(dataset_path):
        return str(dataset_path)

    # Otherwise there may be a sibling file with appended suffix '.agdc-md.yaml'.
    expected_name = dataset_path.parent.joinpath('{}.agdc-md'.format(dataset_path.name))
    found = _find_any_metadata_suffix(expected_name)
    if found:
        return str(found)

    # Otherwise if it's a directory, there may be an 'agdc-metadata.yaml' file describing all contained datasets.
    if dataset_path.is_dir():
        expected_name = dataset_path.joinpath('agdc-metadata')
        found = _find_any_metadata_suffix(expected_name)
        if found:
            return str(found)

    if is_supported_document_type(dataset_path):
        raise ValueError(f'No such file {dataset_path}')
    else:
        raise ValueError(f'No supported metadata docs found for dataset {dataset_path}')


def _find_any_metadata_suffix(path: Path) -> Optional[Path]:
    """
    Find any supported metadata files that exist with the given file path stem.
    (supported suffixes are tried on the name)

    Eg. searching for '/tmp/ga-metadata' will find if any files such as '/tmp/ga-metadata.yaml' or
    '/tmp/ga-metadata.json', or '/tmp/ga-metadata.yaml.gz' etc that exist: any suffix supported by read_documents()

    :type path: pathlib.Path
    """
    existing_paths = list(filter(is_supported_document_type, path.parent.glob(path.name + '*')))
    if not existing_paths:
        return None

    if len(existing_paths) > 1:
        raise ValueError('Multiple matched metadata files: {!r}'.format(existing_paths))

    return existing_paths[0]


def ui_path_doc_stream(paths, logger=None, uri=True, raw=False):
    """Given a stream of URLs, or Paths that could be directories, generate a stream of
    (path, doc) tuples.

    For every path:
    1. If directory find the metadata file or log error if not found

    2. Load all documents from that path and return one at a time (parsing
    errors are logged, but processing should continue)

    :param paths: Filesystem paths

    :param logger: Logger to use to report errors

    :param uri: If True return path in uri format, else return it as filesystem path

    :param raw: By default docs are wrapped in :class:`SimpleDocNav`, but you can
    instead request them to be raw dictionaries

    """

    def on_error1(p, e):
        if logger is not None:
            logger.error(str(e))

    def on_error2(p, e):
        if logger is not None:
            logger.error('Failed reading documents from %s', str(p))

    yield from _path_doc_stream(_resolve_doc_files(paths, on_error=on_error1),
                                on_error=on_error2, uri=uri, raw=raw)


def _resolve_doc_files(paths, on_error):
    for p in paths:
        try:
            yield get_metadata_path(p)
        except ValueError as e:
            on_error(p, e)


def _path_doc_stream(files, on_error, uri=True, raw=False):
    """See :func:`ui_path_doc_stream` for documentation"""
    maybe_wrap = identity if raw else SimpleDocNav

    for fname in files:
        try:
            for p, doc in read_documents(fname, uri=uri):
                yield p, maybe_wrap(doc)

        except InvalidDocException as e:
            on_error(fname, e)



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2020 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
import pytest

import datacube
from integration_tests.test_end_to_end import INGESTER_CONFIGS
from integration_tests.test_full_ingestion import (check_open_with_api,
                                                   ensure_datasets_are_indexed, check_data_shape,
                                                   check_grid_mapping, check_cf_compliance, check_attributes,
                                                   check_dataset_metadata_in_storage_unit,
                                                   check_open_with_xarray)
from integration_tests.utils import prepare_test_ingestion_configuration
import netCDF4


# Current formulation of this test relies on non-EO3 test data
@pytest.mark.parametrize('datacube_env_name', ('datacube', ))
@pytest.mark.timeout(20)
@pytest.mark.usefixtures('default_metadata_type',
                         'indexed_ls5_scene_products')
def test_index_out_of_bound_error(clirunner, index, tmpdir, example_ls5_dataset_paths, ingest_configs):
    """
    Test for the case where ingestor processes upto `--queue-size` number of tasks and not all the available scenes
    """
    # Make a test ingestor configuration
    config = INGESTER_CONFIGS / ingest_configs['ls5_nbar_albers']
    config_path, config = prepare_test_ingestion_configuration(tmpdir, None,
                                                               config, mode='fast_ingest')

    def index_dataset(path):
        return clirunner(['dataset', 'add', str(path)])

    # Set the queue size to process 5 tiles
    queue_size = 5
    valid_uuids = []
    for uuid, ls5_dataset_path in example_ls5_dataset_paths.items():
        valid_uuids.append(uuid)
        index_dataset(ls5_dataset_path)

    # Ensure that datasets are actually indexed
    ensure_datasets_are_indexed(index, valid_uuids)

    # Locationless scenario within database arises when we run the sync tool (with --update-location option)
    # on the disk where the actual file is removed and regenerated again with new dataset id.
    for indexed_uuid in valid_uuids:
        dc1 = datacube.Datacube(index=index)
        datasets = dc1.find_datasets(product='ls5_nbar_scene')
        try:
            # Remove location from the index, to simulate indexed out of range scenario
            res = dc1.index.datasets.remove_location(indexed_uuid, datasets[0].local_uri)
        except AttributeError:
            # Do for one dataset, ignore any other attribute errors
            pass
        assert res is True, "Error for %r. output: %r" % (indexed_uuid, res)

    # Ingest scenes with locationless dataset
    clirunner([
        'ingest',
        '--config-file',
        str(config_path),
        '--queue-size',
        queue_size,
        '--allow-product-changes',
    ])

    # Validate that the ingestion is working as expected
    datasets = index.datasets.search_eager(product='ls5_nbar_albers')
    assert len(datasets) > 0
    assert datasets[0].managed

    check_open_with_api(index, len(valid_uuids))

    # NetCDF specific checks, based on the saved NetCDF file
    ds_path = str(datasets[0].local_path)
    with netCDF4.Dataset(ds_path) as nco:
        check_data_shape(nco)
        check_grid_mapping(nco)
        check_cf_compliance(nco)
        check_dataset_metadata_in_storage_unit(nco, example_ls5_dataset_paths)
        check_attributes(nco, config['global_attributes'])

        name = config['measurements'][0]['name']
        check_attributes(nco[name], config['measurements'][0]['attrs'])
    check_open_with_xarray(ds_path)



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2020 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
"""
Methods for managing dynamic dataset field indexes and views.
"""

import logging

from sqlalchemy import Index
from sqlalchemy import select

from ._core import schema_qualified
from .sql import pg_exists, CreateView

_LOG = logging.getLogger(__name__)

# TODO: Dynamic indexes currently disable in postgis driver.


def contains_all(d_, *keys):
    """
    Does the dictionary have values for all of the given keys?

    >>> contains_all({'a': 4}, 'a')
    True
    >>> contains_all({'a': 4, 'b': 5}, 'a', 'b')
    True
    >>> contains_all({'b': 5}, 'a')
    False
    """
    return all([d_.get(key) for key in keys])


def _ensure_view(conn, fields, name, replace_existing, where_expression):
    """
    Ensure a view exists for the given fields
    """
    view_name = schema_qualified('dv_{}_dataset'.format(name))
    _LOG.warning("Skipping dynamic Index %s ", view_name)


def _old_ensure_view(conn, fields, name, replace_existing, where_expression):
    # Create a view of search fields (for debugging convenience).
    # 'dv_' prefix: dynamic view. To distinguish from views that are created as part of the schema itself.
    view_name = schema_qualified('dv_{}_dataset'.format(name))
    exists = pg_exists(conn, view_name)
    # This currently leaves a window of time without the views: it's primarily intended for development.
    if exists and replace_existing:
        _LOG.debug('Dropping view: %s (replace=%r)', view_name, replace_existing)
        conn.execute('drop view %s' % view_name)
        exists = False
    if not exists:
        _LOG.debug('Creating view: %s', view_name)
        # old_query = select(
        #         [field.alchemy_expression.label(field.name) for field in fields.values()
        #          if not field.affects_row_selection]
        #     ).select_from(
        #         Dataset.join(Product).join(MetadataType)
        #     ).where(where_expression)
        conn.execute(
            CreateView(
                view_name,
                select(
                    [field.alchemy_expression for field in fields.values()
                     if not field.affects_row_selection]
                ).where(where_expression)
            )
        )
    else:
        _LOG.debug('View exists: %s (replace=%r)', view_name, replace_existing)
    legacy_name = schema_qualified('{}_dataset'.format(name))
    if pg_exists(conn, legacy_name):
        _LOG.debug('Dropping legacy view: %s', legacy_name)
        conn.execute('drop view %s' % legacy_name)


def check_dynamic_fields(conn, concurrently, dataset_filter, excluded_field_names, fields, name,
                         rebuild_indexes=False, rebuild_view=False):
    """
    Check that we have expected indexes and views for the given fields
    """
    _LOG.warning("Skipping all dynamic indexes for %s", name)
    return


def old_check_dynamic_fields(conn, concurrently, dataset_filter, excluded_field_names, fields, name,
                             rebuild_indexes=False, rebuild_view=False):
    # If this type has time/space fields, create composite indexes (as they are often searched together)
    # We will probably move these into product configuration in the future.
    composite_indexes = (
        ('lat', 'lon', 'time'),
        ('time', 'lat', 'lon'),
        ('sat_path', 'sat_row', 'time')
    )

    all_exclusions = tuple(excluded_field_names)
    for composite_names in composite_indexes:
        # If all of the fields are available in this product, we'll create a composite index
        # for them instead of individual indexes.
        if contains_all(fields, *composite_names):
            all_are_excluded = set(excluded_field_names) >= set(composite_names)
            _check_field_index(
                conn,
                [fields.get(f) for f in composite_names],
                name, dataset_filter,
                concurrently=concurrently,
                replace_existing=rebuild_indexes,
                # If all fields were excluded individually it should be removed.
                should_exist=not all_are_excluded,
                index_type='gist'
            )
            all_exclusions += composite_names

    # Create indexes for the individual fields.
    for field in fields.values():
        if not field.postgres_index_type:
            continue
        _check_field_index(
            conn, [field],
            name, dataset_filter,
            should_exist=field.indexed and (field.name not in all_exclusions),
            concurrently=concurrently,
            replace_existing=rebuild_indexes,
        )
    # A view of all fields
    _ensure_view(conn, fields, name, rebuild_view, dataset_filter)


def _check_field_index(conn, fields, name_prefix, filter_expression,
                       should_exist=True, concurrently=False,
                       replace_existing=False, index_type=None):
    """
    Check the status of a given index: add or remove it as needed
    """
    if index_type is None:
        if len(fields) > 1:
            raise ValueError('Must specify index type for composite indexes.')
        index_type = fields[0].postgres_index_type

    field_name = '_'.join([f.name.lower() for f in fields])
    # Our normal indexes start with "ix_", dynamic indexes with "dix_"
    index_name = 'dix_{prefix}_{field_name}'.format(
        prefix=name_prefix.lower(),
        field_name=field_name
    )
    # Previous naming scheme
    legacy_name = 'dix_field_{prefix}_dataset_{field_name}'.format(
        prefix=name_prefix.lower(),
        field_name=field_name,
    )
    indexed_expressions = [f.alchemy_expression for f in fields]
    index = Index(
        index_name,
        *indexed_expressions,
        postgresql_where=filter_expression,
        postgresql_using=index_type,
        # Don't lock the table (in the future we'll allow indexing new fields...)
        postgresql_concurrently=concurrently
    )
    exists = pg_exists(conn, schema_qualified(index_name))
    legacy_exists = pg_exists(conn, schema_qualified(legacy_name))

    # This currently leaves a window of time without indexes: it's primarily intended for development.
    if replace_existing or (not should_exist):
        if exists:
            _LOG.debug('Dropping index: %s (replace=%r)', index_name, replace_existing)
            index.drop(conn)
            exists = False
        if legacy_exists:
            _LOG.debug('Dropping legacy index: %s (replace=%r)', legacy_name, replace_existing)
            Index(legacy_name, *indexed_expressions).drop(conn)
            legacy_exists = False

    if should_exist:
        if not (exists or legacy_exists):
            _LOG.info('Creating index: %s', index_name)
            index.create(conn)
        else:
            _LOG.debug('Index exists: %s  (replace=%r)', index_name, replace_existing)
